from __future__ import annotations

import builtins
import contextlib
import enum
import inspect
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator, Mapping
from dataclasses import dataclass, replace
from functools import partial, reduce
from typing import Any, Optional, get_type_hints, override

from greyhorse.maybe import Maybe
from greyhorse.utils.types import (
    TypeWrapper,
    is_awaitable,
    is_maybe,
    is_optional,
    unwrap_maybe,
    unwrap_optional,
)


type FactoryFn[T] = (
    Callable[..., T]
    | Callable[..., Awaitable[T]]
    | type[T]
    | T
    | Callable[..., Generator[T, T, None]]
    | Callable[..., AsyncGenerator[T, T]]
)


class CachePolicy(enum.IntEnum):
    VOLATILE = enum.auto(0)
    GRAPH = enum.auto()
    ANY = enum.auto()


class ParamOpt(int, enum.Enum):
    NORMAL = 0
    OPTIONAL = 1
    MAYBE = 2


@dataclass(slots=True, frozen=True, kw_only=True)
class ParamData:
    type: type
    name: str
    optional: ParamOpt
    value: Any | None = None

    @property
    def is_required(self) -> bool:
        return self.optional == ParamOpt.NORMAL

    @property
    def raw_type(self) -> builtins.type:
        match self.optional:
            case ParamOpt.NORMAL:
                return self.type
            case ParamOpt.OPTIONAL:
                return Optional[self.type]  # noqa: UP045
            case ParamOpt.MAYBE:
                return Maybe[self.type]

    @classmethod
    def from_kv(cls, k: str, v: type, value: Any | None = None) -> ParamData:
        opt = (
            ParamOpt.OPTIONAL
            if is_optional(v)
            else ParamOpt.MAYBE
            if is_maybe(v)
            else ParamOpt.NORMAL
        )
        return cls(type=unwrap_maybe(unwrap_optional(v)), name=k, optional=opt, value=value)


class Factory[T](TypeWrapper[T], ABC):
    __slots__ = ('_cache_policy', '_name_map', '_params', '_type_map')

    scoped: bool = False
    is_async: bool = False

    def __init__(
        self,
        params: Mapping[str, type],
        args: Mapping[str, Any] | None = None,
        cache_policy: CachePolicy = CachePolicy.VOLATILE,
    ) -> None:
        args = args or {}
        params_data = [ParamData.from_kv(k, v, args.get(k)) for k, v in params.items()]
        self._params: list[ParamData] = params_data
        self._name_map = {pd.name: i for i, pd in enumerate(self._params)}
        self._type_map = {pd.type: i for i, pd in enumerate(self._params)}
        self._cache_policy = cache_policy

    @property
    def cache_policy(self) -> CachePolicy:
        return self._cache_policy

    @property
    def cacheable(self) -> bool:
        return self._cache_policy > 0

    @abstractmethod
    def create[**P](self, *args: P.args, **kwargs: P.kwargs) -> T | Awaitable[T]: ...

    @abstractmethod
    def destroy(self, instance: T) -> None | Awaitable[None]: ...

    def __call__[**P](self, *args: P.args, **kwargs: P.kwargs) -> T | Awaitable[T]:
        return self.create(*args, **kwargs)

    def clone(self) -> Factory[T]:
        instance = type(self)({})
        self._clone_into(instance)
        return instance

    @abstractmethod
    def __eq__(self, other: Factory[T]) -> bool: ...

    @abstractmethod
    def __hash__(self) -> int: ...

    @property
    def return_type(self) -> type[T]:
        return self.__wrapped_type__

    @property
    def params_count(self) -> int:
        return len(self._params)

    # @property
    # def params_types(self) -> tuple[type, ...]:
    #     return tuple(pd.raw_type for pd in self._params if pd.value is None)

    @property
    def args_count(self) -> int:
        return reduce(lambda a, pd: a + (1 if pd.value is not None else 0), self._params, 0)

    @property
    def all_params(self) -> Mapping[str, type]:
        res = {}
        for pd in self._params:
            res[pd.name] = pd.raw_type
        return res

    @property
    def actual_params(self) -> Mapping[str, type]:
        res = {}
        for pd in self._params:
            if pd.value is None:
                res[pd.name] = pd.raw_type
        return res

    def check_signature(self, signature: type[Callable]) -> bool:
        sig_ret_type = signature.__args__[-1]
        sig_params = signature.__args__[0 : len(signature.__args__) - 1]

        if not issubclass(self.return_type, sig_ret_type):
            return False

        sig_params_idx = 0

        for pd in self._params:
            if sig_params_idx < len(sig_params):
                sig_param_type = sig_params[sig_params_idx]
                fit = False

                match pd.optional:
                    case ParamOpt.NORMAL:
                        if inspect.isclass(sig_param_type):
                            fit = issubclass(sig_param_type, pd.type)
                        else:
                            fit = sig_param_type is pd.type
                    case ParamOpt.OPTIONAL:
                        fit = issubclass(unwrap_optional(sig_param_type), pd.type)
                    case ParamOpt.MAYBE:
                        fit = issubclass(unwrap_maybe(sig_param_type), pd.type)

                if fit:
                    sig_params_idx += 1
                    continue

            if pd.value is None and pd.optional == ParamOpt.NORMAL:
                return False

        return sig_params_idx == len(sig_params)

    def has_named_param(self, name: str) -> bool:
        return name in self._name_map

    def has_typed_param[U](self, key: type[U]) -> bool:
        return key in self._type_map

    def add_named_arg[U](self, name: str, value: U) -> bool:
        if name not in self._name_map:
            return False
        idx = self._name_map[name]
        if self._params[idx].value is value:
            return False
        # if self._params[idx].optional == _ParamOpt.MAYBE:
        #     value = Maybe(value)
        self._params[idx] = replace(self._params[idx], value=value)
        return True

    def remove_named_arg(self, name: str) -> bool:
        if name not in self._name_map:
            return False
        idx = self._name_map[name]
        self._params[idx] = replace(self._params[idx], value=None)
        return True

    def add_typed_arg[U](self, key: type[U], value: U) -> bool:
        if key not in self._type_map:
            return False
        idx = self._type_map[key]
        if self._params[idx].value is value:
            return False
        # if self._params[idx].optional == _ParamOpt.MAYBE:
        #     value = Maybe(value)
        self._params[idx] = replace(self._params[idx], value=value)
        return True

    def remove_typed_arg[U](self, key: type[U]) -> bool:
        if key not in self._type_map:
            return False
        idx = self._type_map[key]
        self._params[idx] = replace(self._params[idx], value=None)
        return True

    def _clone_into(self, other: Factory[T]) -> None:
        other._params = self._params.copy()
        other._name_map = self._name_map.copy()
        other._type_map = self._type_map.copy()
        other._cache_policy = self._cache_policy

    def _prepare_kwargs[**P](self, *args: P.args, **kwargs: P.kwargs) -> dict[str, Any]:
        prepared_kwargs = {}
        args_idx = 0

        for pd in self._params:
            if args_idx < len(args):
                arg_value = args[args_idx]
                arg_type = type(arg_value)
                fit = False

                match pd.optional:
                    case ParamOpt.NORMAL:
                        fit = issubclass(arg_type, pd.type)
                    case ParamOpt.OPTIONAL:
                        fit = issubclass(unwrap_optional(arg_type), pd.type)
                    case ParamOpt.MAYBE:
                        fit = issubclass(unwrap_maybe(arg_type), pd.type)
                        arg_value = Maybe(arg_value)

                if fit:
                    prepared_kwargs[pd.name] = arg_value
                    args_idx += 1
                    continue

            match pd.optional:
                case ParamOpt.NORMAL:
                    if pd.name in kwargs:
                        prepared_kwargs[pd.name] = kwargs[pd.name]
                    elif pd.value is not None:
                        prepared_kwargs[pd.name] = pd.value
                case ParamOpt.OPTIONAL:
                    prepared_kwargs[pd.name] = kwargs.get(pd.name, pd.value)
                case ParamOpt.MAYBE:
                    if pd.name in kwargs:
                        prepared_kwargs[pd.name] = Maybe(kwargs[pd.name])
                    else:
                        prepared_kwargs[pd.name] = Maybe(pd.value)

        return prepared_kwargs

    @classmethod
    def from_syncfn(
        cls,
        fn: Callable[..., T],
        cache_policy: CachePolicy | None = None,
        hints: Mapping[str, Any] | None = None,
    ) -> Factory[T]:
        if not hints:
            hints = get_type_hints(strip_partial(fn), localns={'T': cls.__wrapped_type__})
        return_type = hints.pop('return', cls.__wrapped_type__)
        kwargs = {'cache_policy': cache_policy} if cache_policy is not None else {}
        return _SyncFnFactory[return_type](fn, hints, **kwargs)

    @classmethod
    def from_asyncfn(
        cls,
        fn: Callable[..., Awaitable[T]],
        cache_policy: CachePolicy | None = None,
        hints: Mapping[str, Any] | None = None,
    ) -> Factory[T]:
        if not hints:
            hints = get_type_hints(strip_partial(fn), localns={'T': cls.__wrapped_type__})
        return_type = hints.pop('return', cls.__wrapped_type__)
        kwargs = {'cache_policy': cache_policy} if cache_policy is not None else {}
        return _AsyncFnFactory[return_type](fn, hints, **kwargs)

    @classmethod
    def from_class(
        cls,
        cls_: type[T],
        cache_policy: CachePolicy | None = None,
        hints: Mapping[str, Any] | None = None,
    ) -> Factory[T]:
        if not hints:
            hints = get_type_hints(cls_.__init__, localns={'T': cls.__wrapped_type__})
        hints.pop('return', None)
        kwargs = {'cache_policy': cache_policy} if cache_policy is not None else {}
        return _ClassFactory[cls_](cls_, hints, **kwargs)

    @classmethod
    def from_instance(cls, instance: T) -> Factory[T]:
        instance_type = type(instance)
        return _InstanceFactory[instance_type](instance)

    @classmethod
    def from_syncgen(
        cls,
        gen_fn: Callable[..., Generator[T, T, None]],
        cache_policy: CachePolicy | None = None,
        hints: Mapping[str, Any] | None = None,
    ) -> Factory[T]:
        if not hints:
            hints = get_type_hints(strip_partial(gen_fn), localns={'T': cls.__wrapped_type__})
        return_type = hints.pop('return', cls.__wrapped_type__)
        if hasattr(return_type, '__origin__') and hasattr(return_type, '__args__'):
            match return_type.__origin__.__name__:
                case 'Generator' | 'Iterator' | 'Iterable':
                    return_type = return_type.__args__[0]
        kwargs = {'cache_policy': cache_policy} if cache_policy is not None else {}
        return _SyncGenFactory[return_type](gen_fn, hints, **kwargs)

    @classmethod
    def from_asyncgen(
        cls,
        gen_fn: Callable[..., AsyncGenerator[T, T]],
        cache_policy: CachePolicy | None = None,
        hints: Mapping[str, Any] | None = None,
    ) -> Factory[T]:
        if not hints:
            hints = get_type_hints(strip_partial(gen_fn), localns={'T': cls.__wrapped_type__})
        return_type = hints.pop('return', cls.__wrapped_type__)
        if hasattr(return_type, '__origin__') and hasattr(return_type, '__args__'):
            match return_type.__origin__.__name__:
                case 'AsyncGenerator' | 'AsyncIterator' | 'AsyncIterable':
                    return_type = return_type.__args__[0]
        kwargs = {'cache_policy': cache_policy} if cache_policy is not None else {}
        return _AsyncGenFactory[return_type](gen_fn, hints, **kwargs)


def into_factory[T](
    fn: FactoryFn[T],
    return_hint: type[T] | None = None,
    cache_policy: CachePolicy | None = None,
) -> Factory[T]:
    orig_fn = strip_partial(fn)
    return_hint = return_hint if return_hint is not None else type

    if callable(fn):
        unwrapped_fn = inspect.unwrap(orig_fn)

        if inspect.isgeneratorfunction(unwrapped_fn):
            return Factory[return_hint].from_syncgen(fn, cache_policy)
        if inspect.isasyncgenfunction(unwrapped_fn):
            return Factory[return_hint].from_asyncgen(fn, cache_policy)
        if inspect.isfunction(orig_fn) or inspect.ismethod(orig_fn):
            if is_awaitable(orig_fn):
                return Factory[return_hint].from_asyncfn(fn, cache_policy)
            return Factory[return_hint].from_syncfn(fn, cache_policy)
    if inspect.isclass(orig_fn):
        return Factory[return_hint].from_class(fn, cache_policy)
    return Factory[return_hint].from_instance(fn)


def strip_partial[T](fn: FactoryFn[T]) -> FactoryFn[T]:
    orig_fn = fn
    while isinstance(orig_fn, partial):
        orig_fn = orig_fn.func
    return orig_fn


class _SyncFnFactory[T](Factory[T]):
    __slots__ = ('_fn',)

    def __init__(
        self,
        fn: Callable[..., T],
        params: Mapping[str, type],
        args: Mapping[str, Any] | None = None,
        cache_policy: CachePolicy = CachePolicy.VOLATILE,
    ) -> None:
        super().__init__(params, args, cache_policy)
        self._fn = fn

    @override
    def create[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        prepared_kwargs = self._prepare_kwargs(*args, **kwargs)
        return self._fn(**prepared_kwargs)

    @override
    def destroy(self, instance: T) -> None:
        del instance

    @override
    def clone(self) -> Factory[T]:
        instance = type(self)(self._fn, {})
        self._clone_into(instance)
        return instance

    @override
    def __eq__(self, other: Factory[T]) -> bool:
        return isinstance(other, _SyncFnFactory) and self._fn == other._fn

    @override
    def __hash__(self) -> int:
        return hash(self._fn)


class _AsyncFnFactory[T](Factory[T]):
    __slots__ = ('_fn',)

    is_async = True

    def __init__(
        self,
        fn: Callable[..., Awaitable[T]],
        params: Mapping[str, type],
        args: Mapping[str, Any] | None = None,
        cache_policy: CachePolicy = CachePolicy.VOLATILE,
    ) -> None:
        super().__init__(params, args, cache_policy)
        self._fn = fn

    @override
    async def create[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        prepared_kwargs = self._prepare_kwargs(*args, **kwargs)
        return await self._fn(**prepared_kwargs)

    @override
    async def destroy(self, instance: T) -> None:
        del instance

    @override
    async def __call__[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        return await self.create(*args, **kwargs)

    @override
    def clone(self) -> Factory[T]:
        instance = type(self)(self._fn, {})
        self._clone_into(instance)
        return instance

    @override
    def __eq__(self, other: Factory[T]) -> bool:
        return isinstance(other, _AsyncFnFactory) and self._fn == other._fn

    @override
    def __hash__(self) -> int:
        return hash(self._fn)


class _ClassFactory[T](Factory[T]):
    __slots__ = ('_cls',)

    def __init__(
        self,
        cls: type[T],
        params: Mapping[str, type],
        args: Mapping[str, Any] | None = None,
        cache_policy: CachePolicy = CachePolicy.GRAPH,
    ) -> None:
        super().__init__(params, args, cache_policy)
        self._cls = cls

    @override
    def create[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        prepared_kwargs = self._prepare_kwargs(*args, **kwargs)
        return self._cls(**prepared_kwargs)

    @override
    def destroy(self, instance: T) -> None:
        del instance

    @override
    def clone(self) -> Factory[T]:
        instance = type(self)(self._cls, {})
        self._clone_into(instance)
        return instance

    @override
    def __eq__(self, other: Factory[T]) -> bool:
        return isinstance(other, _ClassFactory) and self._cls == other._cls

    @override
    def __hash__(self) -> int:
        return hash(self._cls)


class _InstanceFactory[T](Factory[T]):
    __slots__ = ('_instance',)

    def __init__(self, instance: T) -> None:
        super().__init__({}, None, CachePolicy.ANY)
        self._instance = instance

    @override
    def create[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self._instance

    @override
    def destroy(self, instance: T) -> None: ...

    @override
    def clone(self) -> Factory[T]:
        instance = type(self)(self._instance)
        self._clone_into(instance)
        return instance

    @override
    def __eq__(self, other: Factory[T]) -> bool:
        return isinstance(other, _InstanceFactory) and self._instance == other._instance

    @override
    def __hash__(self) -> int:
        return hash(self._instance)


class _SyncGenFactory[T](Factory[T]):
    __slots__ = ('_gen_fn', '_gens')

    scoped = True

    def __init__(
        self,
        generator_fn: Callable[..., Generator[T, T, None]],
        params: Mapping[str, type],
        args: Mapping[str, Any] | None = None,
        cache_policy: CachePolicy = CachePolicy.VOLATILE,
    ) -> None:
        super().__init__(params, args, cache_policy)
        self._gen_fn = generator_fn
        self._gens: dict[int, Generator[T, T, None]] = {}

    @override
    def create[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        prepared_kwargs = self._prepare_kwargs(*args, **kwargs)
        gen = self._gen_fn(**prepared_kwargs)

        try:
            instance = next(gen)

        except StopIteration as e:
            instance = e.value

        self._gens[id(instance)] = gen
        return instance

    @override
    def destroy(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopIteration):
            gen.send(instance)

    @override
    def clone(self) -> Factory[T]:
        instance = type(self)(self._gen_fn, {})
        self._clone_into(instance)
        return instance

    @override
    def __eq__(self, other: Factory[T]) -> bool:
        return isinstance(other, _SyncGenFactory) and self._gen_fn == other._gen_fn

    @override
    def __hash__(self) -> int:
        return hash(self._gen_fn)


class _AsyncGenFactory[T](Factory[T]):
    __slots__ = ('_gen_fn', '_gens')

    scoped = True
    is_async = True

    def __init__(
        self,
        generator_fn: Callable[..., AsyncGenerator[T, T]],
        params: Mapping[str, type],
        args: Mapping[str, Any] | None = None,
        cache_policy: CachePolicy = CachePolicy.VOLATILE,
    ) -> None:
        super().__init__(params, args, cache_policy)
        self._gen_fn = generator_fn
        self._gens: dict[int, AsyncGenerator[T, T]] = {}

    @override
    async def create[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        prepared_kwargs = self._prepare_kwargs(*args, **kwargs)
        gen = self._gen_fn(**prepared_kwargs)

        try:
            instance = await anext(gen)

        except StopAsyncIteration:
            instance = None

        self._gens[id(instance)] = gen
        return instance

    @override
    async def destroy(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopAsyncIteration):
            await gen.asend(instance)

    @override
    async def __call__[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        return await self.create(*args, **kwargs)

    @override
    def clone(self) -> Factory[T]:
        instance = type(self)(self._gen_fn, {})
        self._clone_into(instance)
        return instance

    @override
    def __eq__(self, other: Factory[T]) -> bool:
        return isinstance(other, _AsyncGenFactory) and self._gen_fn == other._gen_fn

    @override
    def __hash__(self) -> int:
        return hash(self._gen_fn)



from __future__ import annotations

from collections import deque
from collections.abc import Mapping
from dataclasses import dataclass, field
from enum import IntEnum
from types import TracebackType
from typing import Any, Self, override

from greyhorse.app.abc.functional.registries import FactoryRegistry
from greyhorse.app.abc.functional.resolver import ResolvedData, Resolver, ResolveResult
from greyhorse.factory import CachePolicy
from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.utils.types import is_maybe, is_optional, unwrap

from ..runtime.invoke import invoke_sync
from .bucket import Bucket


@dataclass(slots=True, kw_only=True)
class _FrameData[T]:
    type: type[T]
    values: dict[type, Any] = field(default_factory=dict)
    visited: bool = False


class BucketResolver(Resolver):
    __slots__ = ('_bucket', '_filters', '_registry')

    def __init__(
        self, registry: FactoryRegistry, prev_bucket: Bucket | None = None, /, **filters: Any
    ) -> None:
        self._registry = registry
        self._filters = filters
        self._bucket = Bucket(prev=prev_bucket)

    @override
    def can_resolve[T](self, key: type[T]) -> bool:
        resolving_type = unwrap(key)
        bucket = self._bucket
        filters = dict(**self._filters)

        while bucket is not None:
            if bucket.cache.has(resolving_type):
                return True
            if bucket.scope is not None:
                filters['scope'] = bucket.scope
            if self._registry.has_factory(target_type=resolving_type, **filters):
                return True
            bucket = bucket.prev

        return False

    @override
    def resolve[T](self, key: type[T], from_scope: IntEnum | None = None) -> Maybe[T]:
        resolving_type = unwrap(key)
        if res := self._bucket.get(resolving_type).map(
            lambda v: Just(v) if is_maybe(key) else v
        ):
            return res

        res = self.resolve_factories(key, from_scope)
        if res.unresolved:
            return Nothing

        return self.resolve_value(res.resolved, key, from_scope)

    @override
    def __enter__(self) -> Self:
        self._bucket.incr()
        return self

    @override
    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self._bucket.decr()

    @override
    def resolve_factories[T](
        self, key: type[T], from_scope: IntEnum | None = None
    ) -> ResolveResult:
        resolve_queue: deque[type] = deque([unwrap(key)])
        resolved: dict[type, ResolvedData] = {}
        unresolved = []

        while resolve_queue:
            resolving_type = resolve_queue[0]
            current_bucket = self._bucket

            while current_bucket is not None:
                if from_scope is None or (
                    current_bucket.scope is not None and current_bucket.scope <= from_scope
                ):
                    if current_bucket.scope is not None:
                        filters = {'scope': current_bucket.scope, **self._filters}
                    else:
                        filters = self._filters

                    if res := self._registry.get_factory(target_type=resolving_type, **filters):
                        resolved_data = ResolvedData(type=resolving_type, factory=res.unwrap())
                        resolved[resolving_type] = resolved_data
                        resolve_queue.popleft()

                        for (
                            param_name,
                            param_type,
                        ) in resolved_data.factory.actual_params.items():
                            stripped_param_type = unwrap(param_type)
                            resolved_data.deps[param_name] = param_type
                            if param_type not in resolved:
                                resolve_queue.append(stripped_param_type)
                        break

                current_bucket = current_bucket.prev
            else:
                unresolved.append(resolving_type)
                resolve_queue.popleft()

        return ResolveResult(resolved=resolved, unresolved=unresolved)

    @override
    def resolve_value[T](  # noqa: PLR0915
        self,
        resolved: Mapping[type, ResolvedData],
        key: type[T],
        from_scope: IntEnum | None = None,
    ) -> Maybe[T]:
        resolving_type = unwrap(key)

        if res := self._bucket.get(resolving_type).map(
            lambda v: Just(v) if is_maybe(key) else v
        ):
            return res

        if resolving_type not in resolved:
            return Nothing

        resolve_stack: list[ResolvedData] = [resolved[resolving_type]]
        frame_stack: list[_FrameData] = [_FrameData(type=resolving_type)]
        instance = None

        while resolve_stack:
            head = resolve_stack[-1]
            frame_head = frame_stack[-1]

            if res := self._bucket.get(head.type):
                frame_head.values[head.type] = res if is_maybe(head.type) else res.unwrap()
                resolve_stack.pop()
                continue

            if head.type is not frame_head.type:
                frame_head = _FrameData(type=head.type)
                frame_stack.append(frame_head)

            if head.deps and not frame_head.visited:
                need_continue = False
                for dep_type in reversed(head.deps.values()):
                    if dep_type not in frame_head.values:
                        stripped_dep_type = unwrap(dep_type)
                        if stripped_dep_type in resolved:
                            resolve_stack.append(resolved[stripped_dep_type])
                            need_continue = True
                        elif is_maybe(dep_type):
                            frame_head.values[dep_type] = Nothing
                        elif is_optional(dep_type):
                            frame_head.values[dep_type] = None
                frame_head.visited = True
                if need_continue:
                    continue

            kwargs = {k: frame_head.values[v] for k, v in head.deps.items()}
            instance = invoke_sync(head.factory.create, **kwargs)
            instance = Just(instance) if is_maybe(head.type) else instance
            if len(frame_stack) > 1:
                frame_stack[-2].values[head.type] = instance

            match head.factory.cache_policy:
                case CachePolicy.VOLATILE | CachePolicy.GRAPH:
                    self._bucket.cache.add(head.type, instance)
                    if head.factory.scoped:
                        self._bucket.finalizers[head.type].append((head.factory, instance))
                case CachePolicy.ANY:
                    bucket = prev_bucket = self._bucket
                    while bucket is not None:
                        if from_scope is None or (
                            bucket.scope is not None and bucket.scope >= from_scope
                        ):
                            bucket.cache.add(head.type, instance)
                            prev_bucket = bucket
                        bucket = bucket.prev
                    if head.factory.scoped:
                        prev_bucket.finalizers[head.type].append((head.factory, instance))

            frame_stack.pop()
            resolve_stack.pop()

        return Just(instance)



from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Mapping
from typing import Any, override

from greyhorse.maybe import Maybe, Nothing


class Selector[K, T](ABC):
    @abstractmethod
    def has(self, key: K) -> bool: ...

    @abstractmethod
    def get(self, key: K) -> Maybe[T]: ...

    @abstractmethod
    def get_with_metadata(self, key: K) -> Maybe[tuple[T, Mapping[str, Any]]]: ...


class ListSelector[K, T](Selector[K, T], ABC):
    @abstractmethod
    def list(self, key: K | None = None) -> Iterable[tuple[K, T] | T]: ...

    @abstractmethod
    def list_with_metadata(
        self, key: K | None = None
    ) -> Iterable[tuple[K, T, Mapping[str, Any]] | tuple[T, Mapping[str, Any]]]: ...

    @abstractmethod
    def filter(
        self, filter_fn: Callable[[K, Mapping[str, Any]], bool] | None = None
    ) -> Iterable[tuple[K, T, Mapping[str, Any]]]: ...


class SelectorCombiner[K, T](Selector[K, T]):
    __slots__ = ('_selectors',)

    def __init__(self, *selectors: Selector[K, T]) -> None:
        self._selectors = list(selectors)

    def __len__(self) -> int:
        return len(self._selectors)

    def add_selector(self, selector: Selector[K, T]) -> bool:
        if selector in self._selectors:
            return False
        self._selectors.append(selector)
        return True

    def remove_selector(self, selector: Selector[K, T]) -> bool:
        if selector not in self._selectors:
            return False
        self._selectors.remove(selector)
        return True

    @override
    def has(self, key: K) -> bool:
        for selector in self._selectors:
            if res := selector.has(key):
                return res
        return False

    @override
    def get(self, key: K) -> Maybe[T]:
        for selector in self._selectors:
            if res := selector.get(key):
                return res
        return Nothing

    @override
    def get_with_metadata(self, key: K) -> Maybe[tuple[T, Mapping[str, Any]]]:
        for selector in self._selectors:
            if res := selector.get_with_metadata(key):
                return res
        return Nothing



from abc import ABC, abstractmethod
from collections.abc import Callable, Collection, Mapping
from typing import Any, overload

from greyhorse.factory import Factory
from greyhorse.maybe import Maybe


class FactoryRegistry(ABC):
    @overload
    def has_factory[T](
        self, target_type: type[T] = ..., signature: None = None, **filters: Any
    ) -> bool: ...

    @overload
    def has_factory(
        self, target_type: None = None, signature: type[Callable] = ..., **filters: Any
    ) -> bool: ...

    @abstractmethod
    def has_factory[T](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        **filters: Any,
    ) -> bool: ...

    @overload
    def get_factory[T](
        self, target_type: type[T] = ..., signature: None = None, **filters: Any
    ) -> Maybe[Factory[T]]: ...

    @overload
    def get_factory(
        self, target_type: None = None, signature: type[Callable] = ..., **filters: Any
    ) -> Maybe[Factory[Any]]: ...

    @abstractmethod
    def get_factory[T](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        **filters: Any,
    ) -> Maybe[Factory[T]]: ...

    @overload
    def get_factory_with_metadata[T](
        self, target_type: type[T] = ..., signature: None = None, **filters: Any
    ) -> Maybe[tuple[Factory[T], Mapping[str, Any]]]: ...

    @overload
    def get_factory_with_metadata(
        self, target_type: None = None, signature: type[Callable] = ..., **filters: Any
    ) -> Maybe[tuple[Factory[Any], Mapping[str, Any]]]: ...

    @abstractmethod
    def get_factory_with_metadata[T](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        **filters: Any,
    ) -> Maybe[tuple[Factory[T], Mapping[str, Any]]]: ...


class MutFactoryRegistry(FactoryRegistry, ABC):
    @abstractmethod
    def add_factory[T](
        self, target_type: type[T], factory: Factory[T], **filters: Any
    ) -> bool: ...

    @abstractmethod
    def remove_factory[T](
        self, target_type: type[T], factory: Factory[T] | None = None, **filters: Any
    ) -> Collection[tuple[Factory[T], Mapping[str, Any]]]: ...



from __future__ import annotations

import inspect
import typing
import weakref
from collections import defaultdict
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from decimal import Decimal
from enum import IntEnum
from itertools import accumulate
from types import NoneType
from typing import Any, cast, override

# import networkx as nx
# from matplotlib.pyplot import plot
from greyhorse.app.abc.collections.registries import Registry
from greyhorse.app.abc.collections.selectors import Selector, SelectorCombiner
from greyhorse.app.abc.resources.common import ResourceEventKind
from greyhorse.app.abc.resources.providers import ProviderListenerFn, ProviderSubscription
from greyhorse.app.private.collections.registries import DictRegistry, MutDictRegistry
from greyhorse.app.private.runtime.invoke import invoke_sync
from greyhorse.factory import CachePolicy, Factory, into_factory
from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.utils.types import is_maybe, is_optional, unwrap_maybe, unwrap_optional


def test_domain_service_factory() -> None:
    ...
    # def f1(a: int)->float:
    #     return float(a)
    #
    # def f11(a0: int)->float:
    #     return float(a0) + 10
    #
    # def f2(a: float)->Decimal:
    #     return Decimal(a)
    #
    # def f3(a: int)->str:
    #     return str(a)
    #
    # f1f = into_factory(f1)
    # f2f = into_factory(f2)
    # f3f = into_factory(f3)
    #
    # assert compositor.add(float, f1f, domain='1', scope=TestLifetime.DOMAIN)
    # assert compositor.add(float, into_factory(f11), domain='1', scope=TestLifetime.OPERATION)
    # assert compositor.add(Decimal, f2f, domain='1', scope=TestLifetime.DOMAIN)
    # assert compositor.add(Decimal, f2f, domain='2', scope=TestLifetime.DOMAIN)
    # assert compositor.add(str, f3f, domain='1', scope=TestLifetime.DOMAIN)
    # assert compositor.add(int, into_factory(lambda: 999, int), domain='1', scope=TestLifetime.DOMAIN)
    # assert compositor.add(int, into_factory(lambda: 999, int), domain='2', scope=TestLifetime.OPERATION)
    #
    # assert compositor.has_provider_signature(Callable[[float], Decimal], domain='1', scope=TestLifetime.DOMAIN)
    # assert compositor.has_provider_signature(Callable[[int], Decimal], domain='1', scope=TestLifetime.DOMAIN)
    # assert compositor.has_provider_signature(Callable[[int], Decimal], domain='2', scope=TestLifetime.OPERATION)

    # fff = compositor.get_provider(Decimal, domain='2', scope=TestLifetime.DOMAIN).unwrap()
    # vvv = fff(123.0)

    # state_ctx = StateContext(TestLifetime, compositor, domain='1')
    #
    # with state_ctx as sel:
    #     assert sel.has(float)
    #     res1 = sel.get(float).unwrap()
    #
    #     with state_ctx as sel1:
    #         assert sel1.has(float)
    #         res2 = sel1.get(float).unwrap()
    #
    #         with state_ctx as sel2:
    #             assert sel2.has(float)
    #             res3 = sel2.get(float).unwrap()
    #
    # assert compositor.remove(float, f1f, domain='1')
    # assert compositor.remove(Decimal, f2f, domain='2')
    # assert compositor.remove(str, f3f, domain='1')

    # assert fragment.add_provider(Decimal, into_factory(Decimal(123)), scope=ModuleScope.MODULE)

    # def decimal_op(value: Decimal) -> int:
    #     return int(value.to_integral_value())

    # assert fragment.add_provider(int, into_factory(decimal_op))

    # frag_ctx = fragment.get_context()

    # assert fragment.has_factory(float)

    # ctx = domain.get_factory_ctx(float).unwrap()
    #
    # with ctx as fn:
    #     assert fn.check_signature(Callable[[], float])
    #     assert fn.check_signature(Callable[[Decimal], float])
    #     fff = fn()
    #     assert float(7890.0) == fff
    #     fn.destroy(fff)
    #
    # ctx = domain.get_factory_ctx(str).unwrap()
    #
    # with ctx as fn:
    #     assert fn.check_signature(Callable[[], str])
    #     assert fn.check_signature(Callable[[Decimal], str])
    #     assert fn.check_signature(Callable[[float], str])
    #     assert fn.check_signature(Callable[[float, Decimal], str])
    #     fff = fn()
    #     assert '579' == fff
    #     fn.destroy(fff)

    # builder = ContextBuilder[SyncContext, TestDataClass](TestDataClass)
    # op = domain.get_provider(float).unwrap()
    # builder.add_param('f1', op)
    # op = domain.get_provider(str).unwrap()
    # builder.add_param('f2', op)
    # ctx = builder.build()
    #
    # with ctx as ddd:
    #     assert ddd.f1.check_signature(Callable[[], float])
    #     assert ddd.f2.check_signature(Callable[[float], str])
    #     composite_f = SyncCompositeFactory[str](ddd.f1, ddd.f2, scoped=True)
    #     assert composite_f.check_signature(Callable[[], str])
    #     v = composite_f()
    #     composite_f.destroy(v)


_Source = typing.NewType('Source', NoneType)
_Sink = typing.NewType('Sink', NoneType)


@dataclass(slots=True, frozen=True, kw_only=True)
class _FactorySchedule[S: IntEnum]:
    factory: Factory[Any]
    policy: CachePolicy
    scoped: bool
    is_async: bool
    scope: S | None


class FactoryCompositor(MutFactoryRegistry, ProviderSubscription):
    __slots__ = ('_deps', '_graph', '_listeners')

    def __init__(self) -> None:
        self._graph = nx.MultiDiGraph()
        self._deps: dict[str, set[type]] = defaultdict(set)
        self._listeners = MutDictRegistry[type, ProviderListenerFn](allow_many=True)

    def clear(self) -> None:
        self._deps.clear()
        self._graph.clear()

    @override
    def has_factory[T, S: IntEnum](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        object_selector: Selector[type, Any] | None = None,
        domain: str = '',
        scope: S | None = None,
        **filters: Any,
    ) -> bool:
        view = self._get_search_view(domain, scope)

        if target_type is None:
            sig_ret_type = signature.__args__[-1]
            for ret_type in view.nodes:
                if ret_type in (_Source, _Sink):
                    continue
                if issubclass(ret_type, sig_ret_type):
                    target_type = ret_type
                    break
            else:
                return False

        if not view.has_node(target_type):
            return False

        schedule = self._schedule(view, target_type, scope)

        if not schedule:
            return False
        if len(schedule) == 1:
            if signature is None:
                return True
            return schedule[0].factory.check_signature(signature)
        factories: list[Factory[Any]] = [f.factory for f in schedule]
        return self._check_signature(factories, signature, object_selector)

    @override
    def get_factory[T, S: IntEnum](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        object_cache: Registry[type, Any] | None = None,
        domain: str = '',
        scope: S | None = None,
        **filters: Any,
    ) -> Maybe[Factory[T]]:
        view = self._get_search_view(domain, scope)

        if target_type is None:
            sig_ret_type = signature.__args__[-1]
            for ret_type in view.nodes:
                if ret_type in (_Source, _Sink):
                    continue
                if issubclass(ret_type, sig_ret_type):
                    target_type = ret_type
                    break
            else:
                return Nothing

        if not view.has_node(target_type):
            return Nothing

        schedule = self._schedule(view, target_type, scope)

        if not schedule:
            return Nothing
        if len(schedule) == 1:
            if signature is None or schedule[0].factory.check_signature(signature):
                return Just(schedule[0].factory.clone())
            return Nothing

        factories: list[Factory[Any]] = [f.factory for f in schedule]

        if signature is not None and not self._check_signature(
            factories, signature, object_cache
        ):
            return Nothing

        cache_policy = max(CachePolicy.STATIC, *[f.policy for f in schedule])
        scoped: bool = any([f.scoped for f in schedule])
        # is_async: bool = any([f.is_async for f in schedule])

        factory = SyncCompositeFactory[target_type](
            *factories, scoped=scoped, object_cache=object_cache, cache_policy=cache_policy
        )

        return Just(factory)

    @override
    def get_factory_with_metadata[T, S: IntEnum](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        object_cache: Registry[type, Any] | None = None,
        domain: str = '',
        scope: S | None = None,
        **filters: Any,
    ) -> Maybe[tuple[Factory[T], Mapping[str, Any]]]:
        return self.get_factory(
            target_type=target_type,
            signature=signature,
            object_cache=object_cache,
            domain=domain,
            scope=scope,
            **filters,
        ).map(lambda f: (f, {}))

    @override
    def add_factory[T, S: IntEnum](
        self,
        target_type: type[T],
        factory: Factory[T],
        domain: str = '',
        scope: S | None = None,
        **metadata: Any,
    ) -> bool:
        if not issubclass(factory.return_type, target_type):
            return False

        node_metadata = {'domains': {domain}}
        edge_metadata = {'domains': {domain}, 'scope': scope}

        if target_type is None:
            self._graph.add_node(_Sink)
            target_type = _Sink

        elif self._graph.has_node(target_type):
            data = self._graph.nodes[target_type]
            if 'domains' not in data:
                data['domains'] = set()
            data['domains'].add(domain)
        else:
            self._graph.add_node(target_type, **node_metadata)

        if target_type in self._deps[domain]:
            self._deps[domain].remove(target_type)

        if not factory.actual_params:
            self._graph.add_edge(_Source, target_type, factory=factory, **edge_metadata)
        else:
            for param_type in factory.actual_params.values():
                stripped_param_type = unwrap_maybe(unwrap_optional(param_type))
                if not self._graph.has_node(stripped_param_type):
                    self._deps[domain].add(stripped_param_type)

                self._graph.add_edge(
                    stripped_param_type, target_type, factory=factory, **edge_metadata
                )

        event = ProviderEvent(
            kind=ResourceEventKind.SETUP, target=factory, type=target_type, metadata=metadata
        )
        filter = _CovariantListenersFilter(kind=ResourceEventKind.SETUP, key=target_type)

        for _, listener, __ in self._listeners.filter(filter):
            listener(event)

        return True

    @override
    def remove_factory[T, S: IntEnum](
        self,
        target_type: type[T],
        factory: Factory[T] | None = None,
        domain: str = '',
        scope: S | None = None,
        **metadata: Any,
    ) -> bool:
        def filter_nodes(in_type: type) -> bool:
            data = self._graph.nodes[in_type]
            return domain in data['domains'] if 'domains' in data else True

        if factory is None:

            def filter_edges(in_type: type, out_type: type, k: int) -> bool:
                data = self._graph[in_type][out_type][k]
                res = domain in data['domains']
                if scope is None:
                    res &= data['scope'] is None
                else:
                    res &= (data['scope'] or 0) <= scope
                return res and out_type is target_type

        else:

            def filter_edges(in_type: type, out_type: type, k: int) -> bool:
                data = self._graph[in_type][out_type][k]
                res = domain in data['domains']
                if scope is None:
                    res &= data['scope'] is None
                else:
                    res &= (data['scope'] or 0) <= scope
                return res and data.get('factory') == factory

        view = cast(
            nx.MultiDiGraph,
            nx.subgraph_view(self._graph, filter_node=filter_nodes, filter_edge=filter_edges),
        )

        if not view.has_node(target_type):
            return Nothing

        old_predecessors = list(view.pred[target_type])
        edges_to_remove = list(view.edges.data('factory', keys=True))
        factories_to_remove = {f for _1, _2, _3, f in edges_to_remove}
        self._graph.remove_edges_from(edges_to_remove)
        has_successors = len(view.succ[target_type]) > 0

        if has_successors:
            self._deps[domain].add(target_type)
        else:
            domains = self._graph.nodes[target_type]['domains']
            domains.remove(domain)
            if not domains:
                self._graph.remove_node(target_type)
            if target_type in self._deps[domain]:
                self._deps[domain].remove(target_type)

        for pred_type in old_predecessors:
            if len(view.succ[pred_type]) == 0:
                self._graph.remove_node(pred_type)
                if pred_type in self._deps[domain]:
                    self._deps[domain].remove(pred_type)

        filter = _CovariantListenersFilter(kind=ResourceEventKind.TEARDOWN, key=target_type)

        for orig_fn in factories_to_remove:
            event = ProviderEvent(
                kind=ResourceEventKind.TEARDOWN,
                target=orig_fn,
                type=target_type,
                metadata=metadata,
            )

            for _, listener, __ in self._prov_listeners.filter(filter):
                listener(event)

        return True

    @override
    def add_provider_listener[T](
        self,
        key: type[T],
        fn: ProviderListenerFn,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool:
        return self._listeners.add(key, fn, kind=kind)

    @override
    def remove_provider_listener[T](
        self,
        key: type[T],
        fn: ProviderListenerFn | None = None,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool:
        filter = _CovariantListenersFilter(kind=kind, key=key)
        to_remove = []

        for t, orig_fn, metadata in self._listeners.filter(filter):
            if fn is None or fn == orig_fn:
                to_remove.append((t, orig_fn, metadata['kind']))

        res = False

        for key, orig_fn, orig_kind in to_remove:
            new_kind = orig_kind & ~kind
            res |= self._listeners.remove(key, orig_fn)
            if new_kind != ResourceEventKind(0):
                self._listeners.add(key, orig_fn, kind=new_kind)

        return res

    def _get_search_view[S: IntEnum](
        self, domain: str = '', scope: S | None = None
    ) -> nx.MultiDiGraph:
        def filter_nodes(in_type: type) -> bool:
            data = self._graph.nodes[in_type]
            return domain in data['domains'] if 'domains' in data else True

        def filter_edges(in_type: type, out_type: type, k: int) -> bool:
            data = self._graph[in_type][out_type][k]
            res = domain in data['domains']
            if scope is None:
                res &= data['scope'] is None
            else:
                res &= (data['scope'] or 0) <= scope
            return res

        return cast(
            nx.MultiDiGraph,
            nx.subgraph_view(self._graph, filter_node=filter_nodes, filter_edge=filter_edges),
        )

    @staticmethod
    def _schedule[T, S: IntEnum](
        view: nx.MultiDiGraph, target_type: type[T], scope: S | None = None
    ) -> list[_FactorySchedule]:
        factories: dict[type, list[_FactorySchedule]] = defaultdict(list)
        type_schedule: list[type] = []
        result: list[_FactorySchedule] = []

        for in_type, out_type, k, _ in nx.edge_bfs(view, target_type, orientation='reverse'):
            data = view.get_edge_data(in_type, out_type, k)
            factory = data['factory']
            factory_scope: S | None = data.get('scope')

            factories[out_type].append(
                _FactorySchedule(
                    factory=factory,
                    policy=factory.cache_policy,
                    scoped=factory.scoped,
                    is_async=factory.is_async,
                    scope=factory_scope,
                )
            )
            if not type_schedule or type_schedule[-1] is not out_type:
                type_schedule.append(out_type)

        for out_type in type_schedule:
            matched: _FactorySchedule | None = None

            for item in factories[out_type]:
                if scope is None and item.scope is None:
                    matched = item
                    break
                if scope is None:
                    if item.scope != 0:
                        continue
                    matched = item
                    break
                if item.scope is None:
                    if scope != 0:
                        continue
                    matched = item
                    break

                if (matched is None and item.scope <= scope) or (
                    matched is not None and min(item.scope, scope) > matched.scope
                ):
                    matched = item

            if matched is None:
                return []
            result.append(matched)

        return result

    @staticmethod
    def _check_signature(
        factories: list[Factory[Any]],
        signature: type[Callable],
        object_selector: Selector[type, Any] | None = None,
    ) -> bool:
        combined_params = []
        has_types = set([f.return_type for f in factories])
        sig_params = signature.__args__[0 : len(signature.__args__) - 1]

        for fn_params in (f.actual_params.values() for f in factories):
            for pt in fn_params:
                if object_selector is not None and object_selector.has(pt):
                    continue
                if pt in has_types and pt not in sig_params:
                    continue
                combined_params.append(pt)

        sig_params_idx = 0

        for param_type in combined_params:
            if sig_params_idx < len(sig_params):
                sig_param_type = sig_params[sig_params_idx]

                if is_maybe(param_type):
                    fit = issubclass(unwrap_maybe(sig_param_type), param_type)
                elif is_optional(param_type):
                    fit = issubclass(unwrap_optional(sig_param_type), param_type)
                elif inspect.isclass(sig_param_type):
                    fit = issubclass(sig_param_type, param_type)
                else:
                    fit = sig_param_type is param_type

                if fit:
                    sig_params_idx += 1
                    continue

            if not is_optional(param_type) and not is_maybe(param_type):
                return False

        return sig_params_idx == len(sig_params)


class SyncCompositeFactory[T](Factory[T]):
    __slots__ = ('_finalizers', '_object_cache', '_schedule', '_scoped')

    def __init__(
        self,
        *factories: Factory[Any],
        scoped: bool,
        args: Mapping[str, Any] | None = None,
        object_cache: DictRegistry[type, Any] | None = None,
        cache_policy: CachePolicy = CachePolicy.VOLATILE,
    ) -> None:
        combined_params = {}
        args = args if args else {}
        has_types = set([f.return_type for f in factories])

        for fn_params in (f._params for f in factories):
            combined_params.update({
                p.name: p.type for p in fn_params if p.type not in has_types
            })
            args.update({
                p.name: p.value
                for p in fn_params
                if p.type not in has_types and p.value is not None
            })

        super().__init__(combined_params, args, cache_policy)
        self._schedule = factories
        self._scoped = scoped
        self._object_cache = object_cache
        self._finalizers: dict[int, list[tuple[Factory[Any], Any]]] = {}

    @property
    def scoped(self) -> bool:
        return self._scoped

    @override
    def create[**P](self, *args: P.args, **kwargs: P.kwargs) -> T:
        actual_args = DictRegistry[type, Any]()

        for param_name, param_value in self._prepare_kwargs(*args, **kwargs).items():
            pd = self._params[self._name_map[param_name]]
            if pd.is_required or param_value is not None:
                actual_args.add(pd.type, param_value)

        factory_cache = DictRegistry[type, Factory[Any]]()
        object_cache = MutDictRegistry[type, Any]()
        selectors = (
            [actual_args, object_cache]
            if self._object_cache is None
            else [actual_args, object_cache, self._object_cache]
        )
        cache_selector = SelectorCombiner[type, Any](*selectors)
        finalizers: list[tuple[Factory[Any], Any | weakref.ref]] = []

        res: T = None

        for i, factory in enumerate(self._schedule):
            last_factory = i == len(self._schedule) - 1
            res_type = factory.return_type

            if actual_args.has(res_type):
                continue

            kwargs: dict[str, Any] = {}

            for param_name, param_type in factory.actual_params.items():
                stripped_param_type = unwrap_maybe(unwrap_optional(param_type))

                if param_value := cache_selector.get(stripped_param_type):
                    kwargs[param_name] = (
                        param_value if is_maybe(param_type) else param_value.unwrap()
                    )

                elif factory_cache.has(stripped_param_type):
                    param_factory, param_kwargs = factory_cache.get_with_metadata(
                        stripped_param_type
                    ).unwrap()
                    param_value = invoke_sync(param_factory.create, **param_kwargs)

                    if param_factory.scoped:
                        finalizers.append(self._create_finalizer(param_factory, param_value))
                    if param_factory.cache_policy < CachePolicy.VOLATILE:
                        object_cache.add(stripped_param_type, param_value)
                    if (
                        param_factory.cache_policy < CachePolicy.SCOPED
                        and self._object_cache is not None
                    ):
                        self._object_cache.add(stripped_param_type, param_value)

                    kwargs[param_name] = (
                        Just(param_value) if is_maybe(param_type) else param_value
                    )

                elif is_maybe(param_type):
                    kwargs[param_name] = Nothing
                elif is_optional(param_type):
                    kwargs[param_name] = None

            if not last_factory:
                factory_cache.add(res_type, factory, **kwargs)
                continue

            res = invoke_sync(factory.create, **kwargs)

            if factory.scoped:
                finalizers.append(self._create_finalizer(factory, res))
            if factory.cache_policy < CachePolicy.SCOPED and self._object_cache is not None:
                self._object_cache.add(res_type, res)

        self._finalizers[id(res)] = finalizers
        return res

    @override
    def destroy(self, instance: T) -> None:
        key = id(instance)

        if finalizers := self._finalizers.get(key):
            for factory, instance in reversed(finalizers):
                if isinstance(instance, weakref.ref):
                    instance = instance()
                invoke_sync(factory.destroy, instance)

            del self._finalizers[key]

    @override
    def clone(self) -> Factory[T]:
        instance = type(self)(self._schedule, self._scoped, self._cacheable, {})
        self._clone_into(instance)
        return instance

    @override
    def __eq__(self, other: Factory[T]) -> bool:
        return isinstance(other, SyncCompositeFactory) and self._schedule == other._schedule

    def _create_finalizer(self, factory: Factory, instance: Any) -> tuple[Factory[Any], Any]:
        if hasattr(instance, '__weakref__'):
            return factory, weakref.ref(instance)
        return factory, instance


if __name__ == '__main__':
    compositor = FactoryCompositor()

    def f1(a: int) -> float:
        return float(a)

    def f11(a0: int) -> float:
        return float(a0) + 10

    def f2(a: float) -> Decimal:
        return Decimal(a)

    def f3(a: int) -> str:
        return str(a)

    f1f = into_factory(f1)
    f2f = into_factory(f2)
    f3f = into_factory(f3)

    assert compositor.add(float, f1f, domain='Домен А', scope=TestLifetime.DOMAIN)
    assert compositor.add(
        float, into_factory(f11), domain='Домен Б', scope=TestLifetime.OPERATION
    )
    assert compositor.add(Decimal, f2f, domain='Домен А', scope=TestLifetime.DOMAIN)
    assert compositor.add(Decimal, f2f, domain='Домен Б', scope=TestLifetime.OPERATION)
    assert compositor.add(str, f3f, domain='Домен А', scope=TestLifetime.DOMAIN)

    import matplotlib.pyplot as plt

    # limits = plt.axis("off")
    # plt.autoscale(True)

    G = compositor._graph
    # pos = nx.nx_agraph.graphviz_layout(G)
    pos = nx.bfs_layout(G, int, align='horizontal')
    # pos = nx.multipartite_layout(G, subset_key="domains")
    # pos = nx.spring_layout(G)

    # nx.nx_agraph.view_pygraphviz(G, edgelabel=lambda d: d['scope'].name)

    connectionstyle = [f'arc3,rad={r}' for r in accumulate([0.15] * 4)]
    connectionstyle2 = [f'arc3,rad={r}' for r in accumulate([0.35] * 4)]

    elarge = [
        tuple(edge) for *edge, d in G.edges(data=True) if d['scope'] == TestLifetime.DOMAIN
    ]
    esmall = [
        tuple(edge) for *edge, d in G.edges(data=True) if d['scope'] == TestLifetime.OPERATION
    ]

    plot()
    nx.draw_networkx_nodes(G, pos=pos, node_size=1900)
    nx.draw_networkx_edges(G, pos, edgelist=elarge, width=4, connectionstyle=connectionstyle)
    nx.draw_networkx_edges(
        G,
        pos,
        edgelist=esmall,
        width=4,
        alpha=0.5,
        edge_color='b',
        style='dashed',
        connectionstyle=connectionstyle2,
    )
    nx.draw_networkx_labels(
        G, pos, labels={n: n.__name__ for n in G}, font_size=10, font_family='sans-serif'
    )
    nx.draw_networkx_edge_labels(
        G,
        pos=pos,
        connectionstyle=connectionstyle,
        edge_labels={
            tuple(edge): f'{",".join(d["domains"])}, {d["scope"].name}'
            for *edge, d in G.edges(keys=True, data=True)
        },
        # bbox={"alpha": 0},
        # label_pos=1.3,
        # font_color="blue",
    )

    ax = plt.gca()
    ax.margins(0.20)
    plt.axis('off')
    plt.tight_layout()
    plt.show()



import inspect
import types
from collections.abc import Callable, Collection, Iterable, Mapping
from dataclasses import dataclass, field, replace
from enum import IntEnum
from typing import Any, ClassVar

from greyhorse.app.abc.functional.registries import MutFactoryRegistry
from greyhorse.app.abc.resources.operators import (
    MutOperatorCollector,
    OperatorCollector,
    OperatorData,
)
from greyhorse.app.abc.resources.providers import (
    MutProviderCollector,
    ProviderCollector,
    ProviderData,
)
from greyhorse.factory import CachePolicy, Factory, FactoryFn, into_factory


@dataclass(slots=True, frozen=True, kw_only=True)
class _MemberData:
    target: Factory[Any] | types.FunctionType
    for_type: type | types.UnionType | None = None
    policy: CachePolicy = CachePolicy.ANY
    scope: IntEnum | None = None
    metadata: Mapping[str, Any] = field(default_factory=dict)


@dataclass(slots=True, frozen=True, kw_only=True)
class _OperatorMemberData:
    target: Factory[Any] | types.FunctionType
    for_type: type | types.UnionType | None = None
    external: Collection[type] | None = None
    scope: IntEnum | None = None
    metadata: Mapping[str, Any] = field(default_factory=dict)


def _is_factory_member(member: types.FunctionType) -> bool:
    return hasattr(member, '__factory_data__')


def _is_operator_member(member: types.FunctionType) -> bool:
    return hasattr(member, '__operator_data__')


def _is_provider_member(member: types.FunctionType) -> bool:
    return hasattr(member, '__provider_data__')


def factory(
    for_type: type | types.UnionType | None = None,
    scope: IntEnum | None = None,
    policy: CachePolicy = CachePolicy.ANY,
    **metadata: Any,
) -> Callable[[FactoryFn[Any]], FactoryFn[Any]]:
    def decorator(func: FactoryFn[Any]) -> FactoryFn[Any]:
        func.__factory_data__ = _MemberData(
            target=func, for_type=for_type, policy=policy, scope=scope, metadata=metadata
        )
        return func

    return decorator


def operator(
    for_type: type | types.UnionType | None = None,
    scope: IntEnum | None = None,
    external: Collection[type] | None = None,
    **metadata: Any,
) -> Callable[[FactoryFn[Any]], FactoryFn[Any]]:
    def decorator(func: FactoryFn[Any]) -> FactoryFn[Any]:
        func.__operator_data__ = _OperatorMemberData(
            target=func, for_type=for_type, external=external, scope=scope, metadata=metadata
        )
        return func

    return decorator


def provider(
    for_type: type | types.UnionType | None = None,
    scope: IntEnum | None = None,
    policy: CachePolicy = CachePolicy.ANY,
    **metadata: Any,
) -> Callable[[FactoryFn[Any]], FactoryFn[Any]]:
    def decorator(func: FactoryFn[Any]) -> FactoryFn[Any]:
        func.__provider_data__ = _MemberData(
            target=func, for_type=for_type, policy=policy, scope=scope, metadata=metadata
        )
        return func

    return decorator


class Fragment:
    __slots__ = ('__factories', '__operators', '__providers')
    name: ClassVar[str]
    scope: ClassVar[type[IntEnum] | None] = None

    def __init__(self) -> None:
        self.__factories: list[_MemberData] = []
        self.__operators: list[_OperatorMemberData] = []
        self.__providers: list[_MemberData] = []
        self._init_members()

    def __repr__(self) -> str:
        return f'Fragment "{self.name}"'

    def setup(
        self,
        component: str,
        factory_registry: MutFactoryRegistry,
        op_collector: OperatorCollector,
        prov_collector: ProviderCollector,
    ) -> None:
        for data in self.__factories:
            self._setup_factory(data, factory_registry)

        for data in self.__operators:
            self._setup_operator(data, factory_registry, op_collector, component)

        for data in self.__providers:
            self._setup_provider(data, prov_collector, component)

    def teardown(
        self,
        component: str,
        factory_registry: MutFactoryRegistry,
        op_collector: MutOperatorCollector,
        prov_collector: MutProviderCollector,
    ) -> None:
        for data in reversed(self.__providers):
            self._teardown_provider(data, prov_collector, component)

        for data in reversed(self.__operators):
            self._teardown_operator(data, factory_registry, op_collector, component)

        for data in reversed(self.__factories):
            self._teardown_factory(data, factory_registry)

    def _init_members(self) -> None:
        for _, member in inspect.getmembers(self):
            if _is_factory_member(member):
                data = member.__factory_data__
                data = replace(data, target=into_factory(member, cache_policy=data.policy))
                self.__factories.append(data)

            elif _is_operator_member(member):
                data = member.__operator_data__
                data = replace(data, target=into_factory(member, cache_policy=CachePolicy.ANY))
                self.__operators.append(data)

            elif _is_provider_member(member):
                data = member.__provider_data__
                data = replace(data, target=into_factory(member, cache_policy=data.policy))
                self.__providers.append(data)

    def _get_target_types(self, data: _MemberData) -> Iterable[type]:
        if data.for_type is None:
            target_types = [data.target.return_type]
        elif isinstance(data.for_type, types.UnionType):
            target_types = data.for_type.__args__
        else:
            target_types = [data.for_type]
        return target_types

    def _setup_factory(self, data: _MemberData, factory_registry: MutFactoryRegistry) -> None:
        target_types = self._get_target_types(data)
        scope = data.scope or (self.scope(0) if self.scope is not None else None)

        for target_type in target_types:
            filters = dict(fragment=self.name)
            if scope is not None:
                filters['scope'] = scope

            factory_registry.add_factory(target_type, data.target, **filters)

    def _teardown_factory(
        self, data: _MemberData, factory_registry: MutFactoryRegistry
    ) -> None:
        target_types = self._get_target_types(data)
        scope = data.scope or (self.scope(0) if self.scope is not None else None)

        for target_type in target_types:
            filters = dict(fragment=self.name)
            if scope is not None:
                filters['scope'] = scope

            factory_registry.remove_factory(target_type, data.target, **filters)

    def _setup_operator(
        self,
        data: _OperatorMemberData,
        factory_registry: MutFactoryRegistry,
        op_collector: OperatorCollector,
        component: str,
    ) -> None:
        scope = data.scope or (self.scope(0) if self.scope is not None else None)
        signature = Callable[[*data.target.actual_params.values()], data.target.return_type]

        op_data = OperatorData(
            component=component,
            fragment=self.name,
            target=data.target,
            signature=signature,
            scope=scope,
            external_params=data.external,
            metadata=data.metadata,
        )
        data = _MemberData(
            target=data.target, for_type=data.for_type, scope=data.scope, metadata=data.metadata
        )

        for target_type in self._get_target_types(data):
            op_collector.add(target_type, op_data)

        self._setup_factory(data, factory_registry)

    def _teardown_operator(
        self,
        data: _OperatorMemberData,
        factory_registry: MutFactoryRegistry,
        op_collector: MutOperatorCollector,
        component: str,
    ) -> None:
        scope = data.scope or (self.scope(0) if self.scope is not None else None)
        signature = Callable[[*data.target.actual_params.values()], data.target.return_type]

        op_data = OperatorData(
            component=component,
            fragment=self.name,
            target=data.target,
            signature=signature,
            scope=scope,
            external_params=data.external,
            metadata=data.metadata,
        )
        data = _MemberData(
            target=data.target, for_type=data.for_type, scope=data.scope, metadata=data.metadata
        )

        for target_type in self._get_target_types(data):
            op_collector.remove(target_type, op_data)

        self._teardown_factory(data, factory_registry)

    def _setup_provider(
        self, data: _MemberData, prov_collector: ProviderCollector, component: str
    ) -> None:
        scope = data.scope or (self.scope(0) if self.scope is not None else None)
        signature = Callable[[*data.target.actual_params.values()], data.target.return_type]

        prov_data = ProviderData(
            component=component,
            fragment=self.name,
            signature=signature,
            scope=scope,
            metadata=data.metadata,
        )

        for target_type in self._get_target_types(data):
            prov_collector.add(target_type, prov_data)

    def _teardown_provider(
        self, data: _MemberData, prov_collector: MutProviderCollector, component: str
    ) -> None:
        scope = data.scope or (self.scope(0) if self.scope is not None else None)
        signature = Callable[[*data.target.actual_params.values()], data.target.return_type]

        prov_data = ProviderData(
            component=component,
            fragment=self.name,
            signature=signature,
            scope=scope,
            metadata=data.metadata,
        )

        for target_type in self._get_target_types(data):
            prov_collector.remove(target_type, prov_data)



from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping
from typing import Any, override

from greyhorse.app.abc.collections.registries import ListRegistry, MutListRegistry
from greyhorse.maybe import Just, Maybe, Nothing


class DictRegistry[K, T](ListRegistry[K, T]):
    __slots__ = ('_allow_many', '_storage')

    def __init__(self, allow_many: bool = False) -> None:
        self._storage: dict[K, list[tuple[T, dict[str, Any]]]] = defaultdict(list)
        self._allow_many = allow_many

    def __len__(self) -> int:
        return len(self._storage)

    @override
    def add(self, key: K, instance: T, /, **metadata: Any) -> bool:
        if key in self._storage:
            variants = self._storage[key]
            if not self._allow_many and len(variants) > 0:
                return instance == variants[0]
            if instance in variants:
                return False
            variants.append((instance, metadata))
        else:
            self._storage[key].append((instance, metadata))
        return True

    @override
    def has(self, key: K) -> bool:
        return key in self._storage

    @override
    def get(self, key: K) -> Maybe[T]:
        if variants := self._storage.get(key):
            item = variants[0]
            return Just(item[0])

        return Nothing

    @override
    def get_with_metadata(self, key: K) -> Maybe[tuple[T, Mapping[str, Any]]]:
        if variants := self._storage.get(key):
            item = variants[0]
            return Just(item)

        return Nothing

    @override
    def list(self, key: K | None = None) -> Iterable[tuple[K, T] | T]:
        if key is None:
            for k, v in self._storage.items():
                for item in v:
                    yield k, item[0]
        else:
            if key not in self._storage:
                return

            for item in self._storage[key]:
                yield item[0]

    @override
    def list_with_metadata(
        self, key: K | None = None
    ) -> Iterable[tuple[K, T, Mapping[str, Any]] | tuple[T, Mapping[str, Any]]]:
        if key is None:
            for k, v in self._storage.items():
                for item in v:
                    yield k, *item
        else:
            if key not in self._storage:
                return

            yield from self._storage[key]

    @override
    def filter(
        self, filter_fn: Callable[[K, Mapping[str, Any]], bool] | None = None
    ) -> Iterable[tuple[K, T, Mapping[str, Any]]]:
        for k, v in self._storage.items():
            for item in v:
                if not filter_fn or filter_fn(k, item[1]):
                    yield k, *item


class MutDictRegistry[K, T](DictRegistry[K, T], MutListRegistry[K, T]):
    @override
    def remove(self, key: K, instance: T | None = None) -> bool:
        if key not in self._storage:
            return False

        result = False

        if instance is None:
            del self._storage[key]
            result = True

        else:
            i = 0
            items = self._storage[key]
            while i < len(items):
                v, _ = items[i]
                if v == instance:
                    result = True
                    self._storage[key].pop(i)
                else:
                    i += 1

            if len(self._storage[key]) == 0:
                del self._storage[key]

        return result

    @override
    def clear(self) -> None:
        while key := next(iter(self._storage), None):
            del self._storage[key]


class ScopedDictRegistry[K, T](ListRegistry[K, T]):
    __slots__ = ('_scope_func', '_storage')

    def __init__(
        self, factory: Callable[[], DictRegistry[K, T]], scope_func: Callable[[], str]
    ) -> None:
        super().__init__()
        self._scope_func = scope_func
        self._storage: dict[str, DictRegistry[K, T]] = defaultdict(factory)

    def _get_registry(self) -> DictRegistry[K, T]:
        key = self._scope_func()
        return self._storage[key]

    def __len__(self) -> int:
        registry = self._get_registry()
        return registry.__len__()

    @override
    def add(self, key: K, instance: T, /, **metadata: Any) -> bool:
        registry = self._get_registry()
        return registry.add(key, instance, **metadata)

    @override
    def has(self, key: K) -> bool:
        registry = self._get_registry()
        return registry.has(key)

    @override
    def get(self, key: K) -> Maybe[T]:
        registry = self._get_registry()
        return registry.get(key)

    @override
    def get_with_metadata(self, key: K) -> Maybe[tuple[T, Mapping[str, Any]]]:
        registry = self._get_registry()
        return registry.get_with_metadata(key)

    @override
    def list(self, key: K | None = None) -> Iterable[tuple[K, T] | T]:
        registry = self._get_registry()
        return registry.list(key)

    @override
    def list_with_metadata(
        self, key: K | None = None
    ) -> Iterable[tuple[K, T, Mapping[str, Any]] | tuple[T, Mapping[str, Any]]]:
        registry = self._get_registry()
        return registry.list_with_metadata(key)

    @override
    def filter(
        self, filter_fn: Callable[[K, Mapping[str, Any]], bool] | None = None
    ) -> Iterable[tuple[K, T, Mapping[str, Any]]]:
        registry = self._get_registry()
        yield from registry.filter(filter_fn)


class ScopedMutDictRegistry[K, T](ScopedDictRegistry[K, T], MutListRegistry[K, T]):
    __slots__ = ('_storage',)

    def __init__(
        self, factory: Callable[[], MutDictRegistry[K, T]], scope_func: Callable[[], str]
    ) -> None:
        super().__init__(factory, scope_func)
        self._storage: dict[str, MutDictRegistry[K, T]] = defaultdict(factory)

    def _get_registry(self) -> MutDictRegistry[K, T]:
        key = self._scope_func()
        return self._storage[key]

    @override
    def remove(self, key: K, instance: T | None = None) -> bool:
        registry = self._get_registry()
        return registry.remove(key, instance)

    @override
    def clear(self) -> None:
        registry = self._get_registry()
        registry.clear()
        del self._storage[self._scope_func()]



from abc import ABC, abstractmethod
from typing import Any, override


class Collector[K, T](ABC):
    @abstractmethod
    def add(self, key: K, instance: T, /, **metadata: Any) -> bool: ...


class MutCollector[K, T](Collector[K, T], ABC):
    @abstractmethod
    def remove(self, key: K, instance: T | None = None) -> bool: ...


class CollectorCombiner[K, T](MutCollector[K, T]):
    __slots__ = ('_collectors',)

    def __init__(self, *collectors: MutCollector[K, T]) -> None:
        self._collectors = list(collectors)

    def __len__(self) -> int:
        return len(self._collectors)

    def add_collector(self, collector: MutCollector[K, T]) -> bool:
        if collector in self._collectors:
            return False
        self._collectors.append(collector)
        return True

    def remove_collector(self, collector: MutCollector[K, T]) -> bool:
        if collector not in self._collectors:
            return False
        self._collectors.remove(collector)
        return True

    @override
    def add(self, key: K, instance: T, /, **metadata: Any) -> bool:
        res = False
        for collector in self._collectors:
            res |= collector.add(key, instance, **metadata)
        return res

    @override
    def remove(self, key: K, instance: T | None = None) -> bool:
        res = False
        for collector in self._collectors:
            res |= collector.remove(key, instance)
        return res



import inspect
from collections.abc import Callable, Mapping
from typing import Any

from greyhorse.app.abc.collections.selectors import Selector
from greyhorse.maybe import Nothing

from .types import is_maybe, is_optional, unwrap_maybe, unwrap_optional


TypeProviderFactory = Callable[[str, ...], Any]


class ParamsInjector:
    def __init__(self, type_selector: Selector[type, Any] | None = None) -> None:
        self._type_selector = type_selector

    def __call__(
        self,
        func: Callable[[...], ...],
        values: Mapping[str, Any] | None = None,
        types: Mapping[type, Any] | None = None,
    ) -> inspect.BoundArguments:
        values = values or {}
        types = types or {}

        sig = inspect.signature(func, eval_str=True)
        args = {}

        for k, v in sig.parameters.items():
            if k == 'values':
                args[k] = values
            elif v.annotation in types:
                args[k] = types[v.annotation]
            elif k in values:
                args[k] = values[k]
            elif self._type_selector:
                if value := self._type_selector.get(
                    unwrap_maybe(unwrap_optional(v.annotation))
                ):
                    args[k] = value if is_maybe(v.annotation) else value.unwrap()
                elif is_maybe(v):
                    args[k] = Nothing
                elif is_optional(v):
                    args[k] = None

        args = sig.bind_partial(**args)
        args.apply_defaults()
        return args



from __future__ import annotations

from collections.abc import Collection, MutableMapping
from dataclasses import dataclass
from typing import override

from greyhorse.app.abc.providers import BorrowError, BorrowMutError
from greyhorse.app.private.collections.registries import MutDictRegistry
from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.result import Ok, Result


@dataclass(slots=True, frozen=True, kw_only=True)
class ResourceProvisionPolicy:
    allow_borrow_when_acquired = False
    allow_acq_when_borrowed = False
    allow_multiple_acquisition = False


# type SharedFactory[T] = Factory[T] | Factory[Result[T, BorrowError]]
# type MutFactory[T] = Factory[T] | Factory[Result[T, BorrowMutError]]


class SharedStateImpl[S](ResourceState[S]):
    __slots__ = (
        '_conditions',
        '_next_status',
        '_policy',
        '_shared_counter',
        '_status',
        '_type',
        '_value',
    )

    def __init__(self, type: type[S], policy: ResourceProvisionPolicy | None = None) -> None:
        self._policy: ResourceProvisionPolicy = policy or ResourceProvisionPolicy()
        self._status: ResourceStatus = ResourceStatus.Idle
        self._conditions: MutableMapping[str, Condition] = {}
        self._next_status: ResourceStatus | None = None
        self._shared_counter: int = 0
        self._type = type
        self._value: Maybe[S] = Nothing

    def __repr__(self) -> str:
        return f'ResourceState <{self.shared_type.__name__}> status={self._status}'

    @property
    @override
    def shared_type(self) -> type[S]:
        return self._type

    @property
    @override
    def status(self) -> ResourceStatus:
        return self._status

    @property
    @override
    def transition_status(self) -> TransitionStatus:
        total = len(self._conditions)
        completed = 0
        permanent = False
        reasons = []
        messages = []

        for condition in self._conditions.values():
            if condition.status:
                completed += 1
            else:
                permanent |= condition.permanent
                reasons.append(condition.name)
                if condition.message:
                    messages.append(condition.message)

        if total == completed:
            return TransitionStatus.Complete
        if not permanent:
            if completed == 0:
                return TransitionStatus.Pending
            return TransitionStatus.Partial(total=total, completed=completed)
        return TransitionStatus.Failure(reasons=reasons, message='\n'.join(messages))

    @property
    @override
    def next_status(self) -> ResourceStatus | None:
        return self._next_status

    @override
    def borrow(self) -> Result[S, BorrowError]:
        match self._status:
            case ResourceStatus.Available(alive, _):
                self._status = ResourceStatus.Available(alive=alive, in_use=True)
            case ResourceStatus.Acquired(_):
                if not self._policy.allow_borrow_when_acquired:
                    return BorrowError.BorrowedAsMutable(
                        name=self.shared_type.__name__
                    ).to_result()
            case _:
                return BorrowError.Empty(name=self.shared_type.__name__).to_result()

        if self._value.is_nothing():
            return BorrowError.Empty(name=self.shared_type.__name__).to_result()

        self._shared_counter += 1
        return Ok(self._value.unwrap())

    @override
    def reclaim(self) -> None:
        match self._status:
            case ResourceStatus.Available(alive, _):
                self._status = ResourceStatus.Available(
                    alive=alive, in_use=self._shared_counter > 1
                )
            case ResourceStatus.Acquired(_):
                pass
            case _:
                return

        if self._shared_counter == 1:
            self._shared_counter = 0
            self.reset_value()
        else:
            self._shared_counter = max(0, self._shared_counter - 1)

    def set_condition(self, condition: Condition) -> None:
        self._conditions[condition.name] = condition

    def reset_condition(self, name: str) -> bool:
        return self._conditions.pop(name, None) is not None

    def set_status(self, status: ResourceStatus) -> None:
        self._status = status

    def set_value(self, value: S) -> S:
        self._value = Just(value)
        self._shared_counter = 1
        return value

    def reset_value(self) -> None:
        self._value = Nothing


class MutStateImpl[S, M](SharedStateImpl[S], MutResourceState[S, M]):
    __slots__ = ('_acq_counter', '_mut_type', '_mut_value')

    def __init__(
        self,
        shared_type: type[S],
        mut_type: type[M],
        policy: ResourceProvisionPolicy | None = None,
    ) -> None:
        super().__init__(shared_type, policy)
        self._acq_counter: int = 0
        self._mut_type = mut_type
        self._mut_value: Maybe[M] = Nothing

    def __repr__(self) -> str:
        return f'MutResourceState <{self.shared_type.__name__}, {self.mut_type.__name__}> status={self._status}'

    @property
    @override
    def mut_type(self) -> type[M]:
        return self._mut_type

    @override
    def acquire(self) -> Result[M, BorrowMutError]:
        match self._status:
            case ResourceStatus.Available(alive, in_use):
                if in_use and not self._policy.allow_acq_when_borrowed:
                    return BorrowMutError.BorrowedAsImmutable(
                        name=self.mut_type.__name__
                    ).to_result()
                self._status = ResourceStatus.Acquired(alive=alive)
            case ResourceStatus.Acquired(_):
                if not self._policy.allow_multiple_acquisition:
                    return BorrowMutError.AlreadyBorrowed(
                        name=self.mut_type.__name__
                    ).to_result()
            case _:
                return BorrowMutError.Empty(name=self.mut_type.__name__).to_result()

        if self._mut_value.is_nothing():
            return BorrowMutError.Empty(name=self.mut_type.__name__).to_result()

        self._acq_counter += 1
        return Ok(self._mut_value.unwrap())

    @override
    def release(self) -> None:
        match self._status:
            case ResourceStatus.Acquired(alive):
                if self._acq_counter == 1:
                    self._acq_counter = 0
                    self._status = ResourceStatus.Available(
                        alive=alive, in_use=self._shared_counter > 1
                    )
                    self.reset_mut_value()
                else:
                    self._acq_counter = max(0, self._acq_counter - 1)
            case _:
                return

    def set_mut_value(self, value: M) -> M:
        self._mut_value = Just(value)
        self._acq_counter = 1
        return value

    def reset_mut_value(self) -> None:
        self._mut_value = Nothing


class _ReconcilerTaskHandler(TaskHandler):
    def __init__(
        self,
        task: ReconcilerTask,
        reconciler: ReconcilerImpl,
        allowed_conditions: list[str] | None = None,
    ) -> None:
        self._task = task
        self._reconciler = reconciler
        self._allowed_conditions = allowed_conditions

    @override
    def add_event(self, event: ReconcilerEvent) -> None:
        self._reconciler.add_event(event)

    @override
    def set_condition(self, condition: Condition) -> bool:
        if (
            self._allowed_conditions is not None
            and condition.name not in self._allowed_conditions
        ):
            return False
        self._reconciler.set_condition(condition)
        return True

    @override
    def cancel(self, reason: str | None = None) -> None:
        self._reconciler.on_task_cancel(self._task.name, reason=reason)

    @override
    def complete[T](self, value: T | None = None) -> None:
        self._reconciler.on_task_complete(self._task.name, value=value)

    @override
    def failure(self, message: str, exception: Exception | None = None) -> None:
        self._reconciler.on_task_failure(self._task.name, message=message, exception=exception)


class ReconcilerImpl[S](Reconciler[S]):
    __slots__ = ('_events', '_state', '_type')

    def __init__(
        self, state: SharedStateImpl[S], waiter: ReconcilerWaiter | None = None
    ) -> None:
        self._state = state
        self._waiter = waiter
        self._tasks = MutDictRegistry[str, ReconcilerTask]()
        self._setup_schedule = []
        self._teardown_schedule = []
        self._periodic_schedule = []
        self._permanent_schedule = []
        self._events = []

    def _schedule(self) -> None:
        for _, task, _md in self._tasks.filter(lambda n, md: md['pending'] is True):
            match task.phase:
                case ReconcilerRunPhase.Setup:
                    self._setup_schedule.append(task)
                case ReconcilerRunPhase.Teardown:
                    self._teardown_schedule.append(task)
                case ReconcilerRunPhase.Periodic(interval):
                    self._periodic_schedule.append(task)
                case ReconcilerRunPhase.Permanent(async_):
                    self._permanent_schedule.append(task)

    @property
    @override
    def status(self) -> ReconciliationStatus:
        match self._state.status:
            case ResourceStatus.Available(alive, _) | ResourceStatus.Acquired(alive):
                if alive:
                    return ReconciliationStatus.Live(status=self._state.transition_status)
                return ReconciliationStatus.Succeeded(status=self._state.transition_status)
            case ResourceStatus.Failed(message):
                return ReconciliationStatus.Failed(
                    status=self._state.transition_status, message=message
                )
            case _:
                return ReconciliationStatus.Dead

    @property
    @override
    def waiter(self) -> ReconcilerWaiter | None:
        return self._waiter

    @override
    def set_resource_status(self, status: ResourceStatus) -> None:
        match status:
            case ResourceStatus.Available(alive, _) | ResourceStatus.Acquired(alive):
                if alive:
                    self._waiter.value.clear()
                else:
                    self._waiter.value.set()
            case _:
                pass

        self._state._status = status

    @override
    def set_condition(self, condition: Condition) -> None:
        self._state.set_condition(condition)

    @override
    def reset_condition(self, name: str) -> bool:
        return self._state.reset_condition(name)

    @override
    def events(self) -> Collection[ReconcilerEvent]:
        return self._events

    @override
    def setup(self, task_collector: TaskCollector) -> None: ...

    @override
    def teardown(self) -> None: ...

    def add_event(self, event: ReconcilerEvent) -> None:
        self._events.append(event)

    def on_task_cancel(self, name: str, reason: str | None = None) -> None: ...

    def on_task_complete(self, name: str, value: T | None = None) -> None: ...

    def on_task_failure(
        self, name: str, message: str, exception: Exception | None = None
    ) -> None: ...



from abc import abstractmethod
from collections.abc import Awaitable
from typing import Protocol

from greyhorse.error import Error, ErrorCase
from greyhorse.result import Result

from .resources.borrow import BorrowError, BorrowMutError


class Provider[T](Protocol):
    @abstractmethod
    def borrow(self) -> Result[T, BorrowError] | Awaitable[Result[T, BorrowError]]: ...

    @abstractmethod
    def reclaim(self) -> None | Awaitable[None]: ...


class MutProvider[T](Protocol):
    @abstractmethod
    def acquire(self) -> Result[T, BorrowMutError] | Awaitable[Result[T, BorrowMutError]]: ...

    @abstractmethod
    def release(self) -> None | Awaitable[None]: ...


class ForwardError(Error):
    namespace = 'greyhorse.app'

    Empty = ErrorCase(
        msg='Cannot forward "{name}" because the value is not available', name=str
    )

    MovedOut = ErrorCase(
        msg='Cannot forward "{name}" because the value was moved out', name=str
    )

    Unexpected = ErrorCase(
        msg='Cannot forward "{name}" because an unexpected error occurred: "{details}"',
        name=str,
        details=str,
    )

    # InsufficientDeps = ErrorCase(
    #     msg='Cannot forward "{name}" because dependencies are not enough to satisfy', name=str
    # )


class ForwardProvider[T](Protocol):
    @abstractmethod
    def take(self) -> Result[T, ForwardError] | Awaitable[Result[T, ForwardError]]: ...

    @abstractmethod
    def drop(self, instance: T) -> None | Awaitable[None]: ...

    @abstractmethod
    def __bool__(self) -> bool | Awaitable[bool]: ...


AnyProvider = Provider | MutProvider | ForwardProvider



from __future__ import annotations

from collections import defaultdict
from collections.abc import Mapping
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Any, override

from greyhorse.app.abc.collections.registries import MutRegistry
from greyhorse.app.abc.collections.selectors import Selector
from greyhorse.factory import Factory
from greyhorse.maybe import Maybe, Nothing

from ..collections.registries import MutDictRegistry
from ..runtime.invoke import invoke_sync


@dataclass(slots=True, kw_only=True)
class Bucket(Selector[type, Any]):
    cache: MutRegistry[type, Any] = field(
        default_factory=MutDictRegistry[type, Any], repr=False
    )
    finalizers: dict[type, list[tuple[Factory[Any], Any]]] = field(
        default_factory=lambda: defaultdict(list), repr=False
    )
    prev: Bucket | None = field(default=None, repr=False)
    scope: IntEnum | None = field(default=None, repr=True)
    num_children: int = field(default=0, repr=True)
    _counter: int = field(default=0, init=False, repr=True)

    @override
    def has[T](self, key: type[T]) -> bool:
        bucket = self

        while bucket is not None:
            if bucket.cache.has(key):
                return True
            bucket = bucket.prev
        return False

    @override
    def get[T](self, key: type[T]) -> Maybe[T]:
        bucket = self

        while bucket is not None:
            if (v := bucket.cache.get(key)).is_just():
                return v
            bucket = bucket.prev
        return Nothing

    @override
    def get_with_metadata[T](self, key: type[T]) -> Maybe[tuple[T, Mapping[str, Any]]]:
        bucket = self

        while bucket is not None:
            if (v := bucket.cache.get_with_metadata(key)).is_just():
                return v
            bucket = bucket.prev
        return Nothing

    def incr(self) -> None:
        bucket = self

        while bucket is not None:
            bucket._counter += 1  # noqa: SLF001
            bucket = bucket.prev

    def decr(self) -> None:
        bucket = self

        while bucket is not None:
            bucket._counter -= 1  # noqa: SLF001
            if bucket._counter == 0:  # noqa: SLF001
                bucket._finalize()  # noqa: SLF001
            bucket = bucket.prev

    def _finalize(self) -> None:
        while key := next(reversed(self.finalizers), None):
            for factory, instance in reversed(self.finalizers[key]):
                invoke_sync(factory.destroy, instance)
            del self.finalizers[key]

        self.cache.clear()



from __future__ import annotations

from collections import defaultdict
from collections.abc import AsyncGenerator, Generator, Mapping, Sequence
from dataclasses import dataclass, field
from decimal import Decimal
from enum import IntEnum, auto
from types import TracebackType
from typing import Any, Self, override

from greyhorse.app.abc.collections.selectors import Selector
from greyhorse.app.contexts import CtxCallbacks, SyncCallbackContext, SyncContext
from greyhorse.app.private.collections.registries import MutDictRegistry
from greyhorse.app.private.runtime.invoke import invoke_sync
from greyhorse.factory import Factory, into_factory
from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.utils.types import (
    TypeWrapper,
    is_maybe,
    is_optional,
    unwrap_maybe,
    unwrap_optional,
)


@dataclass(slots=True, frozen=True, kw_only=True)
class _Get[T]:
    instance: T
    cached: bool
    metadata: Mapping[str, Any] = field(default_factory=dict)


class Container[T: IntEnum](Selector[type, Any], TypeWrapper[T]):
    __slots__ = (
        '_cache',
        '_ctx',
        '_current_level',
        '_parent',
        '_resources',
        '_scoped_factories',
        '_selectors',
    )

    def __init__(
        self,
        selectors: Sequence[Selector[type, Factory[Any]]],
        current_level: T | None = None,
        parent: Container | None = None,
        resources: dict[type, Any] | None = None,
    ) -> None:
        self._selectors = selectors
        self._current_level = current_level if current_level else self.wrapped_type(0)
        self._parent = parent
        self._resources = MutDictRegistry[type, Any]()
        self._cache = MutDictRegistry[type, Any]()
        self._scoped_factories: dict[type, list[tuple[Factory[Any], Any]]] = defaultdict(list)

        if resources:
            for k, v in resources.items():
                self._resources.add(k, v)

        self._ctx = self._create_ctx()

    def __repr__(self) -> str:
        return f'Container <{self._current_level.name}>'

    @property
    def parent(self) -> Container | None:
        return self._parent

    @property
    def level(self) -> T:
        return self._current_level

    @property
    def context(self) -> SyncContext[Container]:
        return self._ctx

    @override
    def has[T](self, key: type[T]) -> bool:
        if self._cache.has(key) or self._get_registry().has(key):
            return True
        if self._parent is not None:
            return self._parent.has(key)
        return False

    @override
    def get[T](self, key: type[T]) -> Maybe[T]:
        return self._get(key).map(lambda x: x.instance)

    @override
    def get_with_metadata[T](self, key: type[T]) -> Maybe[tuple[T, Mapping[str, Any]]]:
        return self._get(key).map(lambda x: (x.instance, x.metadata))

    def __call__(
        self, resources: dict[type, Any] | None = None, scope: T | None = None
    ) -> SyncContext[Container]:
        if scope is not None:
            if scope <= self._current_level:
                return self.context
            if scope >= self._current_level + len(self._selectors):
                return self.context

        child = self

        if scope is None:
            for cur_level in range(self._current_level, len(self.wrapped_type) + 1):
                level_idx = cur_level - self._current_level
                if level_idx < 1:
                    continue
                if level_idx < len(self._selectors):
                    child = Container[self.wrapped_type](
                        self._selectors[level_idx:],
                        self.wrapped_type(cur_level),
                        parent=child,
                        resources=resources,
                    )
                    break
                # if not cur_level.autocreate:
                #     break
        else:
            for cur_level in range(self._current_level, len(self.wrapped_type) + 1):
                level_idx = cur_level - self._current_level
                if level_idx < 1:
                    continue
                if child.level >= scope:
                    break
                if level_idx < len(self._selectors):
                    child = Container[self.wrapped_type](
                        self._selectors[level_idx:],
                        self.wrapped_type(cur_level),
                        parent=child,
                        resources=resources,
                    )

        return child.context

    def _get_registry(self) -> Selector[type, Factory[Any]]:
        return self._selectors[0]

    def _get[T](self, key: type[T]) -> Maybe[_Get[T]]:
        if res := self._cache.get_with_metadata(key):
            instance, metadata = res.unwrap()
            return Just(_Get(instance=instance, cached=True, metadata=metadata))

        if res := self._get_registry().get_with_metadata(key):
            factory, metadata = res.unwrap()

            kwargs = {}
            for param_name, param_type in factory.actual_params.items():
                stripped_param_type = unwrap_maybe(unwrap_optional(param_type))

                if res := self._get(stripped_param_type):
                    res = res.unwrap()

                    kwargs[param_name] = (
                        Just(res.instance) if is_maybe(param_type) else res.instance
                    )
                elif is_maybe(param_type):
                    kwargs[param_name] = Nothing
                elif is_optional(param_type):
                    kwargs[param_name] = None

            instance = invoke_sync(factory.create, **kwargs)

            if factory.scoped:
                self._scoped_factories[key].append((factory, instance))
            if factory.cacheable:
                self._cache.add(key, instance, **metadata)

            return Just(_Get(instance=instance, cached=factory.cacheable, metadata=metadata))

        if not self._parent:
            return Nothing

        if not (res := self._parent._get(key)):  # noqa: SLF001
            return Nothing

        res = res.unwrap()
        if res.cached:
            self._cache.add(key, res.instance, **res.metadata)
        return Just(res)

    def _enter(self, _) -> Self:  # noqa: ANN001
        if self._parent:
            self._parent.context.__enter__()

        for k, v in self._resources.list():
            self._cache.add(k, v)

        return self

    def _exit(
        self,
        _,  # noqa: ANN001
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        while key := next(reversed(self._scoped_factories), None):
            for factory, instance in reversed(self._scoped_factories[key]):
                invoke_sync(factory.destroy, instance)
            del self._scoped_factories[key]

        self._cache.clear()

        if self._parent:
            self._parent.context.__exit__(exc_type, exc_value, traceback)

    def _create_ctx(self) -> SyncContext[Container]:
        return SyncCallbackContext[Container](
            callbacks=CtxCallbacks(on_enter=self._enter, on_exit=self._exit),
            factory=lambda *_: self,
        )


def sync_ctx_manager() -> Generator[str, str, None]:
    res = '123'
    yield res
    assert res == '123'


async def async_ctx_manager() -> AsyncGenerator[Decimal, Decimal]:
    res = Decimal(789)
    yield res
    assert res == Decimal(789)


def complex_fn(a: int, b: str) -> Decimal:
    return Decimal(a) + Decimal(b)


class Level(IntEnum):
    ROOT = auto(0)
    FUNCTION = auto()
    LOCAL = auto()


def test_container() -> None:
    registries = [MutDictRegistry[type, Any]() for _ in range(0, len(Level))]

    container = Container[Level](registries)
    assert container.level == Level.ROOT

    assert not container.has(int)
    assert not container.has(str)
    assert not container.has(Decimal)

    assert registries[Level.ROOT].add(int, into_factory(lambda: 123))
    assert registries[Level.ROOT].add(str, into_factory(sync_ctx_manager))

    assert container.has(int)
    assert container.has(str)
    assert not container.has(Decimal)

    assert container.get(int).unwrap() == 123

    with container.context:
        assert container.level == Level.ROOT

        with container() as c1:
            assert c1.level == Level.FUNCTION
            assert registries[Level.FUNCTION].add(Decimal, into_factory(async_ctx_manager))

            assert c1.has(int)
            assert c1.has(str)
            assert c1.has(Decimal)

            assert registries[Level.LOCAL].add(int, into_factory(lambda: 456))

            with c1() as c2:
                assert c2.level == Level.LOCAL
                assert c2.get(int).unwrap() == 456
                assert c2.get(str).unwrap() == '123'
                assert Decimal(789) == c2.get(Decimal).unwrap()

                assert registries[Level.LOCAL].add(Decimal, into_factory(complex_fn))
                assert Decimal(579) == c2.get(Decimal).unwrap()

        assert container.get(int).unwrap() == 123
        assert registries[Level.FUNCTION].remove(Decimal)

    assert registries[Level.ROOT].remove(int)

    assert not container.has(int)
    assert container.has(str)
    assert not container.has(Decimal)



import asyncio
import inspect
import threading
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import Any

from greyhorse.enum import Enum, Struct, Unit
from greyhorse.error import Error, ErrorCase
from greyhorse.result import Result


class ControllerState(Enum):
    Idle = Unit()
    Initialization = Unit()
    Finalization = Unit()
    Initialized = Unit()
    Finalized = Unit()
    Running = Struct(alive=bool)
    Succeeded = Unit()
    Failed = Struct(message=str)


class ControllerWaiter(Enum):
    Sync = Struct(value=threading.Event)
    Async = Struct(value=asyncio.Event)


class ControllerRestartPolicy(Enum):
    Always = Unit()
    Never = Unit()
    OnFailure = Unit()


class ControllerError(Error):
    namespace = 'greyhorse.app'

    Factory = ErrorCase(msg='Controller factory error: "{details}"', details=str)
    NoSuchResource = ErrorCase(msg='Resource "{name}" is not set', name=str)


type ControllerFactoryFn = Callable[[...], Controller | Result[Controller, ControllerError]]
type ControllerFactories = dict[type[Controller], ControllerFactoryFn]


class ControllerVisitor:
    def visit_operator_listener(self, member: OperatorListenerData) -> None: ...

    def visit_provider_listener(self, member: ProviderListenerData) -> None: ...


class Controller(ABC):
    __slots__ = ('__conditions', '__op_listeners', '__prov_listeners', '__state', '__waiter')

    def __init__(self, waiter: ControllerWaiter) -> None:
        self.__state = ControllerState.Idle
        self.__conditions: dict[str, Condition] = {}
        self.__op_listeners: list[OperatorListenerData] = []
        self.__prov_listeners: list[ProviderListenerData] = []
        self.__waiter = waiter
        self._init_members()

    @property
    def state(self) -> ControllerState:
        return self.__state

    @property
    def waiter(self) -> ControllerWaiter:
        return self.__waiter

    @property
    def status(self) -> ReconciliationStatus:
        match self.__state:
            case ControllerState.Running(alive) if alive:
                return ReconciliationStatus.Live(
                    data=ConditionsData(conditions=self.__conditions)
                )
            case ControllerState.Succeeded:
                return ReconciliationStatus.Succeeded(
                    data=ConditionsData(conditions=self.__conditions)
                )
            case ControllerState.Failed:
                return ReconciliationStatus.Failed(
                    data=ConditionsData(conditions=self.__conditions)
                )
            case _:
                return ReconciliationStatus.Dead

    def inspect(self, visitor: ControllerVisitor) -> None:
        for item in self.__op_listeners:
            visitor.visit_operator_listener(item)
        for item in self.__prov_listeners:
            visitor.visit_provider_listener(item)

    @abstractmethod
    def setup(self, *args: Any, **kwargs: Any) -> None | Awaitable[None]: ...

    @abstractmethod
    def teardown(self, *args: Any, **kwargs: Any) -> None | Awaitable[None]: ...

    def _init_members(self) -> None:
        for _, member in inspect.getmembers(self, is_operator_listener_member):
            data = member.__op_list_data__
            member = OperatorListenerData(
                target=member, type=data.type, kind=data.kind, metadata=data.metadata
            )
            self.__op_listeners.append(member)

        for _, member in inspect.getmembers(self, is_provider_listener_member):
            data = member.__prov_list_data__
            member = ProviderListenerData(
                target=member, type=data.type, kind=data.kind, metadata=data.metadata
            )
            self.__prov_listeners.append(member)

    def _set_state(self, state: ControllerState) -> None:
        match state:
            case ControllerState.Running(alive):
                if alive:
                    self.__waiter.value.clear()
                else:
                    self.__waiter.value.set()
            case _:
                pass

        self.__state = state

    def _set_condition(self, condition: Condition) -> None:
        self.__conditions[condition.name] = condition

    def _reset_condition(self, name: str) -> bool:
        return self.__conditions.pop(name, None) is not None



from collections import defaultdict
from collections.abc import Callable, Collection, Mapping
from dataclasses import dataclass, field
from functools import partial
from typing import Any, override

from greyhorse.app.abc.component import Component as BaseComponent
from greyhorse.app.abc.functional.context import OperationContext
from greyhorse.app.abc.resources.operators import (
    MutOperatorCollector,
    OperatorCollector,
    OperatorData,
)
from greyhorse.app.abc.resources.providers import (
    MutProviderCollector,
    ProviderCollector,
    ProviderData,
)
from greyhorse.factory import Factory

from .fragment import Fragment
from .functional.context import BucketOperationContext
from .functional.factory_storage import FactoryStorage
from .functional.operators import ContextOperator


@dataclass(frozen=True, slots=True, kw_only=True)
class _ComponentOperatorCollector(MutOperatorCollector):
    data: dict[str, list[tuple[type, OperatorData]]] = field(
        default_factory=partial(defaultdict, list)
    )

    @override
    def add[U](self, target_type: type[U], instance: OperatorData, /, **metadata: Any) -> bool:
        self.data[instance.fragment].append((target_type, instance))
        return True

    @override
    def remove[U](self, target_type: type[U], instance: OperatorData | None = None) -> bool:
        self.data[instance.fragment].remove((target_type, instance))
        return True


@dataclass(frozen=True, slots=True, kw_only=True)
class _ComponentProviderCollector(MutProviderCollector):
    data: dict[str, list[tuple[type, ProviderData]]] = field(
        default_factory=partial(defaultdict, list)
    )

    @override
    def add[U](self, target_type: type[U], instance: ProviderData, /, **metadata: Any) -> bool:
        self.data[instance.fragment].append((target_type, instance))
        return True

    @override
    def remove[U](self, target_type: type[U], instance: ProviderData | None = None) -> bool:
        self.data[instance.fragment].remove((target_type, instance))
        return True


class Component(BaseComponent):
    __slots__ = (
        '_contexts',
        '_exports',
        '_factories',
        '_fragments',
        '_name',
        '_op_registry',
        '_operators',
        '_prov_registry',
    )

    def __init__(
        self,
        name: str,
        fragments: Collection[Fragment],
        exports: Collection[type] | None = None,
    ) -> None:
        self._name = name
        self._fragments = list(fragments)
        self._factories = FactoryStorage()
        self._operators = FactoryStorage()
        self._contexts = {
            f.name: BucketOperationContext(self._factories, f.scope, fragment=f.name)
            for f in self._fragments
        }
        self._exports = set(exports) if exports else set()
        self._prov_registry = _ComponentProviderCollector()
        self._op_registry = _ComponentOperatorCollector()

    def __repr__(self) -> str:
        return f'Component "{self._name}"'

    @property
    @override
    def name(self) -> str:
        return self._name

    @override
    def context(self) -> OperationContext:
        return BucketOperationContext(self._operators)

    @override
    def add_factory[T](self, target_type: type[T], factory: Factory[T], **filters: Any) -> bool:
        return self._factories.add_factory(target_type, factory, **filters)

    @override
    def remove_factory[T](
        self, target_type: type[T], factory: Factory[T] | None = None, **filters: Any
    ) -> Collection[tuple[Factory[T], Mapping[str, Any]]]:
        return self._factories.remove_factory(target_type, factory, **filters)

    @override
    def setup(
        self,
        op_collector: OperatorCollector | None = None,
        prov_collector: ProviderCollector | None = None,
    ) -> None:
        for fragment in self._fragments:
            fragment.setup(self._name, self._factories, self._op_registry, self._prov_registry)

        for fragment in self._fragments:
            for target_type, prov_data in self._prov_registry.data[fragment.name]:
                if res := self._operators.get_factory(signature=prov_data.signature):
                    self._factories.add_factory(
                        target_type,
                        res.unwrap(),
                        fragment=prov_data.fragment,
                        scope=prov_data.scope,
                        **prov_data.metadata,
                    )
                elif prov_collector is not None:
                    prov_collector.add(target_type, prov_data)

            for target_type, op_data in self._op_registry.data[fragment.name]:
                fragment_context = self._contexts[fragment.name]
                ctx_operator = ContextOperator.from_context(
                    fragment_context, op_data.target, op_data.scope, op_data.external_params
                )
                ctx_operator.compile()
                target = ctx_operator.get_functor()
                self._operators.add_factory(target_type, target)
                if op_collector is not None and target_type in self._exports:
                    signature = Callable[[*target.actual_params.values()], target_type]
                    op_data = OperatorData(
                        component=self._name,
                        fragment=fragment.name,
                        target=target,
                        signature=signature,
                    )
                    op_collector.add(target_type, op_data)

    @override
    def teardown(
        self,
        op_collector: MutOperatorCollector | None = None,
        prov_collector: MutProviderCollector | None = None,
    ) -> None:
        for fragment in reversed(self._fragments):
            for target_type, prov_data in reversed(self._prov_registry.data[fragment.name]):
                if res := self._operators.get_factory(signature=prov_data.signature):
                    self._factories.remove_factory(
                        target_type,
                        res.unwrap(),
                        fragment=prov_data.fragment,
                        scope=prov_data.scope,
                        **prov_data.metadata,
                    )
                elif prov_collector is not None:
                    prov_collector.remove(target_type, prov_data)

            for target_type, op_data in reversed(self._op_registry.data[fragment.name]):
                for target, _md in self._operators.remove_factory(target_type):
                    if op_collector is not None and target_type in self._exports:
                        signature = Callable[[*target.actual_params.values()], target_type]
                        op_data = OperatorData(
                            component=self._name,
                            fragment=fragment.name,
                            target=target,
                            signature=signature,
                        )
                        op_collector.remove(target_type, op_data)

        for fragment in reversed(self._fragments):
            fragment.teardown(
                self._name, self._factories, self._op_registry, self._prov_registry
            )



from typing import override

from greyhorse.app.abc.providers import (
    BorrowError,
    BorrowMutError,
    ForwardError,
    ForwardProvider,
    MutProvider,
    Provider,
)
from greyhorse.app.private.runtime.invoke import invoke_sync
from greyhorse.factory import Factory, FactoryFn, into_factory
from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.result import Ok, Result
from greyhorse.utils.types import TypeWrapper


type SharedFactory[T] = Factory[T] | Factory[Result[T, BorrowError]]
type MutFactory[T] = Factory[T] | Factory[Result[T, BorrowMutError]]


class _BasicBox:
    __slots__ = ('_acq_counter', '_shared_counter')

    allow_borrow_when_acquired = False
    allow_acq_when_borrowed = False
    allow_multiple_acquisition = False

    def __init__(self) -> None:
        self._shared_counter = 0
        self._acq_counter = 0

    def _check_borrow[T](self, name: str) -> BorrowError | None:
        if not self.allow_borrow_when_acquired and self._acq_counter > 0:
            return BorrowError.BorrowedAsMutable(name=name)
        return None

    def _check_borrow_mut[T](self, name: str) -> BorrowMutError | None:
        if not self.allow_multiple_acquisition and self._acq_counter > 0:
            return BorrowMutError.AlreadyBorrowed(name=name)
        if not self.allow_acq_when_borrowed and self._shared_counter > 0:
            return BorrowMutError.BorrowedAsImmutable(name=name)
        return None

    @staticmethod
    def _ensure_result[T](value: object, name: str) -> Result[T, BorrowError]:
        if isinstance(value, Result):
            return value
        if value is None:
            return BorrowError.Empty(name=name).to_result()
        if isinstance(value, BorrowError):
            return value.to_result()
        return Ok(value)

    @staticmethod
    def _ensure_mut_result[T](value: object, name: str) -> Result[T, BorrowMutError]:
        if isinstance(value, Result):
            return value
        if value is None:
            return BorrowMutError.Empty(name=name).to_result()
        if isinstance(value, BorrowMutError):
            return value.to_result()
        return Ok(value)


class SharedBox[T](_BasicBox, Provider[T], TypeWrapper[T]):
    __slots__ = ('_factory', '_value')

    def __init__(self, factory: SharedFactory[T] | FactoryFn[T]) -> None:
        super().__init__()
        if not isinstance(factory, Factory):
            factory = into_factory(factory, self.wrapped_type)
        self._factory = factory
        self._value: Maybe[T] = Nothing

    @override
    def borrow(self) -> Result[T, BorrowError]:
        if (res := self._check_borrow(self.wrapped_type.__name__)) is not None:
            return res.to_result()

        if self._value.is_just():
            self._shared_counter += 1
            return Ok(self._value.unwrap())

        return self._ensure_result(invoke_sync(self._factory), self.wrapped_type.__name__).map(
            self._set_value
        )

    @override
    def reclaim(self) -> None:
        if self._value.is_nothing():
            return
        self._shared_counter = max(0, self._shared_counter - 1)
        if self._shared_counter == 0:
            self._value.map(self._reset_value)

    def _set_value(self, value: T) -> T:
        self._value = Just(value)
        self._shared_counter = 1
        return value

    def _reset_value(self, value: T) -> None:
        if self._factory.scoped:
            invoke_sync(self._factory.destroy, value)
        self._value = Nothing


class MutBox[T](_BasicBox, MutProvider[T], TypeWrapper[T]):
    __slots__ = ('_factory', '_value')

    def __init__(self, factory: MutFactory[T] | FactoryFn[T]) -> None:
        super().__init__()
        if not isinstance(factory, Factory):
            factory = into_factory(factory, self.wrapped_type)
        self._factory = factory
        self._value: Maybe[T] = Nothing

    @override
    def acquire(self) -> Result[T, BorrowMutError]:
        if (res := self._check_borrow_mut(self.wrapped_type.__name__)) is not None:
            return res.to_result()

        if self._value.is_just():
            self._acq_counter += 1
            return Ok(self._value.unwrap())

        return self._ensure_mut_result(
            invoke_sync(self._factory), self.wrapped_type.__name__
        ).map(self._set_value)

    @override
    def release(self) -> None:
        if self._value.is_nothing():
            return
        self._acq_counter = max(0, self._acq_counter - 1)
        if self._acq_counter == 0:
            self._value.map(self._reset_value)

    def _set_value(self, value: T) -> T:
        self._value = Just(value)
        self._acq_counter = 1
        return value

    def _reset_value(self, value: T) -> None:
        if self._factory.scoped:
            invoke_sync(self._factory.destroy, value)
        self._value = Nothing


class CombinedBox[TS, TM](_BasicBox, MutProvider[TM], Provider[TS], TypeWrapper[TS, TM]):
    __slots__ = ('_factory', '_mut_factory', '_mut_value', '_value')

    def __init__(
        self,
        factory: SharedFactory[TS] | FactoryFn[TS],
        mut_factory: MutFactory[TM] | FactoryFn[TM],
    ) -> None:
        super().__init__()
        if not isinstance(factory, Factory):
            factory = into_factory(factory, self.wrapped_type[0])
        if not isinstance(mut_factory, Factory):
            mut_factory = into_factory(mut_factory, self.wrapped_type[1])
        self._factory = factory
        self._mut_factory = mut_factory
        self._value: Maybe[TS] = Nothing
        self._mut_value: Maybe[TM] = Nothing

    @property
    def shared_type(self) -> type[TS]:
        return self._factory.return_type

    @property
    def mut_type(self) -> type[TM]:
        return self._mut_factory.return_type

    @override
    def borrow(self) -> Result[TS, BorrowError]:
        if (res := self._check_borrow(self.shared_type.__name__)) is not None:
            return res.to_result()

        if self._value.is_just():
            self._shared_counter += 1
            return Ok(self._value.unwrap())

        return self._ensure_result(invoke_sync(self._factory), self.shared_type.__name__).map(
            self._set_value
        )

    @override
    def reclaim(self) -> None:
        if self._value.is_nothing():
            return
        self._shared_counter = max(0, self._shared_counter - 1)
        if self._shared_counter == 0:
            self._value.map(self._reset_value)

    @override
    def acquire(self) -> Result[TM, BorrowMutError]:
        if (res := self._check_borrow_mut(self.mut_type.__name__)) is not None:
            return res.to_result()

        if self._mut_value.is_just():
            self._acq_counter += 1
            return Ok(self._mut_value.unwrap())

        return self._ensure_mut_result(
            invoke_sync(self._mut_factory), self.mut_type.__name__
        ).map(self._set_mut_value)

    @override
    def release(self) -> None:
        if self._mut_value.is_nothing():
            return
        self._acq_counter = max(0, self._acq_counter - 1)
        if self._acq_counter == 0:
            self._mut_value.map(self._reset_mut_value)

    def _set_value(self, value: TS) -> TS:
        self._value = Just(value)
        self._shared_counter = 1
        return value

    def _reset_value(self, value: TS) -> None:
        if self._factory.scoped:
            invoke_sync(self._factory.destroy, value)
        self._value = Nothing

    def _set_mut_value(self, value: TM) -> TM:
        self._mut_value = Just(value)
        self._acq_counter = 1
        return value

    def _reset_mut_value(self, value: TM) -> None:
        if self._mut_factory.scoped:
            invoke_sync(self._mut_factory.destroy, value)
        self._mut_value = Nothing


class ForwardBox[T](ForwardProvider[T], TypeWrapper[T]):
    __slots__ = ('_value',)

    def __init__(self, value: T | None = None) -> None:
        self._value = Maybe(value)

    def accept(self, value: T) -> bool:
        if self._value.is_just():
            return False
        self._value = Maybe(value)
        return True

    def revoke(self) -> Maybe[T]:
        value, self._value = self._value, Nothing
        return value

    @override
    def take(self) -> Result[T, ForwardError]:
        value, self._value = self._value, Nothing
        return value.map(Ok).unwrap_or(
            ForwardError.Empty(name=self.wrapped_type.__name__).to_result()
        )

    @override
    def drop(self, instance: T) -> None:
        del instance

    @override
    def __bool__(self) -> bool:
        return self._value.is_just()


class PermanentForwardBox[T](ForwardBox[T]):
    @override
    def take(self) -> Result[T, ForwardError]:
        return self._value.map(Ok).unwrap_or(
            ForwardError.Empty(name=self.wrapped_type.__name__).to_result()
        )

    @override
    def drop(self, instance: T) -> None:
        pass
